"""Simulation class following the PICMI standard
This should be the base classes for Python implementation of the PICMI standard
"""
import math
import sys

from .base import _ClassWithInit

# ---------------------
# Main simulation object
# ---------------------

class PICMI_Simulation(_ClassWithInit):
    """
    Creates a Simulation object

    Parameters
    ----------
    solver: field solver instance
        This is the field solver to be used in the simulation.
        It should be an instance of field solver classes.

    time_step_size: float
        Absolute time step size of the simulation [s].
        Needed if the CFL is not specified elsewhere.

    max_steps: integer
        Maximum number of time steps.
        Specify either this, or `max_time`, or use the `step` function directly.

    max_time: float
        Maximum physical time to run the simulation [s].
        Specify either this, or `max_steps`, or use the `step` function directly.

    verbose: integer, optional
        Verbosity flag. A larger integer results in more verbose output

    particle_shape: {'NGP', 'linear', 'quadratic', 'cubic'}
        Default particle shape for species added to this simulation

    gamma_boost: float, optional
        Lorentz factor of the boosted simulation frame.
        Note that all input values should be in the lab frame.
    """

    def __init__(self, solver=None, time_step_size=None, max_steps=None, max_time=None, verbose=None,
                particle_shape='linear', gamma_boost=None, load_balancing=None, **kw):

        self.solver = solver
        self.time_step_size = time_step_size
        self.verbose = verbose
        self.max_steps = max_steps
        self.max_time = max_time
        self.particle_shape = particle_shape
        self.gamma_boost = gamma_boost

        self.species = []
        self.layouts = []
        self.initialize_self_fields = []
        self.injection_plane_positions = []
        self.injection_plane_normal_vectors = []

        self.lasers = []
        self.laser_injection_methods = []

        self.applied_fields = []

        self.diagnostics = []

        self.interactions = []

        self.load_balancing = load_balancing

        self.handle_init(kw)

    def add_species(self, species, layout, initialize_self_field=None):
        """
        Add species to be used in the simulation

        Parameters
        ----------
        species: species instance
            An instance of one of the PICMI species objects.
            Defines species to be added from the *physical* point of view
            (e.g. charge, mass, initial distribution of particles).

        layout: layout instance
            An instance of one of the PICMI particle layout objects.
            Defines how particles are added into the simulation, from the *numerical* point of view.

        initialize_self_field: bool, optional
            Whether the initial space-charge fields of this species
            is calculated and added to the simulation
        """
        self.species.append(species)
        self.layouts.append(layout)
        self.initialize_self_fields.append(initialize_self_field)
        self.injection_plane_positions.append(None)
        self.injection_plane_normal_vectors.append(None)


    def add_species_through_plane(self, species, layout,
                                  injection_plane_position, injection_plane_normal_vector,
                                  initialize_self_field=None):
        """
        Add species to be used in the simulation that are injected through a plane
        during the simulation.

        Parameters
        ----------
        species: species instance
            An instance of one of the PICMI species objects.
            Defines species to be added from the *physical* point of view
            (e.g. charge, mass, initial distribution of particles).

        layout: layout instance
            An instance of one of the PICMI layout objects.
            Defines how particles are added into the simulation, from the *numerical* point of view.

        initialize_self_field: bool, optional
            Whether the initial space-charge fields of this species
            is calculated and added to the simulation

        injection_plane_position: vector of floats
            Position of one point of the injection plane

        injection_plane_normal_vector: vector of floats
            Vector normal to injection plane
        """
        self.species.append(species)
        self.layouts.append(layout)
        self.initialize_self_fields.append(initialize_self_field)
        self.injection_plane_positions.append(injection_plane_position)
        self.injection_plane_normal_vectors.append(injection_plane_normal_vector)


    def add_laser(self, laser, injection_method):
        """
        Add a laser pulses that to be injected in the simulation

        Parameters
        ----------
        laser_profile: laser instance
            One of laser profile instances.
            Specifies the **physical** properties of the laser pulse
            (e.g. spatial and temporal profile, wavelength, amplitude, etc.).

        injection_method: laser injection instance, optional
            Specifies how the laser is injected (numerically) into the simulation
            (e.g. through a laser antenna, or directly added to the mesh).
            This argument describes an **algorithm**, not a physical object.
            It is up to each code to define the default method
            of injection, if the user does not provide injection_method.
        """
        self.lasers.append(laser)
        self.laser_injection_methods.append(injection_method)

    def add_applied_field(self, applied_field):
        """
        Add an applied field

        Parameters
        ----------
        applied_field: applied field instance
            One of the applied field instance.
            Specifies the properties of the applied field.
        """
        self.applied_fields.append(applied_field)

    def add_diagnostic(self, diagnostic):
        """
        Add a diagnostic

        Parameters
        ----------
        diagnostic: diagnostic instance
            One of the diagnostic instances.
        """
        self.diagnostics.append(diagnostic)

    def add_interaction(self, interaction):
        """
        Add an interaction

        Parameters
        ----------
        interaction: interaction instance
            One of the interaction objects.
        """
        self.interactions.append(interaction)

    def set_max_step(self, max_steps):
        """
        Set the default number of steps for the simulation (i.e. the number
        of steps that gets written when calling `write_input_file`).

        Note: this is equivalent to passing `max_steps` as an argument,
        when initializing the `Simulation` object

        Parameter
        ---------
        max_steps: integer
            Maximum number of time steps
        """
        self.max_steps = max_steps

    def write_input_file(self, file_name):
        """
        Write the parameters of the simulation, as defined in the PICMI input,
        into a code-specific input file.

        This can be used for codes that are not Python-driven (e.g. compiled,
        pure C++ or Fortran codes) and expect a text input in a given format.

        Parameters
        ----------
        file_name: string
            The path to the file that will be created
        """
        raise NotImplementedError

    def step(self, nsteps=1):
        """
        Run the simulation for `nsteps` timesteps

        Parameters
        ----------
        nsteps: integer, default=1
            The number of timesteps
        """
        raise NotImplementedError

    def extension(self):
        """
        Reserved for code-specific extensions, for example returns a class instance
        that has further methods for manipulating a PIC simulation.
        """
        raise NotImplementedError



"""Classes following the PICMI standard
These should be the base classes for Python implementation of the PICMI standard
"""
import re

from .base import _ClassWithInit

# ---------------
# Applied fields
# ---------------


class PICMI_ConstantAppliedField(_ClassWithInit):
    """
    Describes a constant applied field

    Parameters
    ----------
    Ex: float, default=0.
        Constant Ex field [V/m]

    Ey: float, default=0.
        Constant Ey field [V/m]

    Ez: float, default=0.
        Constant Ez field [V/m]

    Bx: float, default=0.
        Constant Bx field [T]

    By: float, default=0.
        Constant By field [T]

    Bz: float, default=0.
        Constant Bz field [T]

    lower_bound: vector, optional
        Lower bound of the region where the field is applied [m].

    upper_bound: vector, optional
        Upper bound of the region where the field is applied [m]
    """
    def __init__(self, Ex=None, Ey=None, Ez=None, Bx=None, By=None, Bz=None,
                 lower_bound=[None,None,None], upper_bound=[None,None,None],
                 **kw):

        self.Ex = Ex
        self.Ey = Ey
        self.Ez = Ez
        self.Bx = Bx
        self.By = By
        self.Bz = Bz

        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

        self.handle_init(kw)


class PICMI_AnalyticAppliedField(_ClassWithInit):
    """
    Describes an analytic applied field

    The expressions should be in terms of the position and time, written as 'x', 'y', 'z', 't'.
    Parameters can be used in the expression with the values given as additional keyword arguments.
    Expressions should be relative to the lab frame.

    Parameters
    ----------
    Ex_expression: string, optional
        Analytic expression describing Ex field [V/m]

    Ey_expression: string, optional
        Analytic expression describing Ey field [V/m]

    Ez_expression: string, optional
        Analytic expression describing Ez field [V/m]

    Bx_expression: string, optional
        Analytic expression describing Bx field [T]

    By_expression: string, optional
        Analytic expression describing By field [T]

    Bz_expression: string, optional
        Analytic expression describing Bz field [T]

    lower_bound: vector, optional
        Lower bound of the region where the field is applied [m].

    upper_bound: vector, optional
        Upper bound of the region where the field is applied [m]
    """
    def __init__(self, Ex_expression=None, Ey_expression=None, Ez_expression=None,
                       Bx_expression=None, By_expression=None, Bz_expression=None,
                 lower_bound=[None,None,None], upper_bound=[None,None,None],
                 **kw):

        self.Ex_expression = Ex_expression
        self.Ey_expression = Ey_expression
        self.Ez_expression = Ez_expression
        self.Bx_expression = Bx_expression
        self.By_expression = By_expression
        self.Bz_expression = Bz_expression

        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

        # --- Find any user defined keywords in the kw dictionary.
        # --- Save them and delete them from kw.
        # --- It's up to the code to make sure that all parameters
        # --- used in the expression are defined.
        self.user_defined_kw = {}
        for k in list(kw.keys()):
            if ((self.Ex_expression is not None and re.search(r'\b%s\b'%k, self.Ex_expression)) or
                (self.Ey_expression is not None and re.search(r'\b%s\b'%k, self.Ey_expression)) or
                (self.Ez_expression is not None and re.search(r'\b%s\b'%k, self.Ez_expression)) or
                (self.Bx_expression is not None and re.search(r'\b%s\b'%k, self.Bx_expression)) or
                (self.By_expression is not None and re.search(r'\b%s\b'%k, self.By_expression)) or
                (self.Bz_expression is not None and re.search(r'\b%s\b'%k, self.Bz_expression))):
                self.user_defined_kw[k] = kw[k]
                del kw[k]

        self.handle_init(kw)


class PICMI_Mirror(_ClassWithInit):
    """
    Describes a perfectly reflecting mirror, where the E and B fields are zeroed
    out in a plane of finite thickness.

    Parameters
    ----------
    x_front_location: float, optional (see comment below)
        Location in x of the front of the nirror [m]

    y_front_location: float, optional (see comment below)
        Location in y of the front of the nirror [m]

    z_front_location: float, optional (see comment below)
        Location in z of the front of the nirror [m]

    depth: float, optional (see comment below)
        Depth of the mirror [m]

    number_of_cells: integer, optional (see comment below)
        Minimum numer of cells zeroed out


    Only one of the [x,y,z]_front_location should be specified. The mirror will be set
    perpendicular to the respective direction and infinite in the others.
    The depth of the mirror will be the maximum of the specified depth and number_of_cells,
    or the code's default value if neither are specified.
    """

    def __init__(self, x_front_location=None, y_front_location=None, z_front_location=None,
                 depth=None, number_of_cells=None, **kw):

        assert [x_front_location,y_front_location,z_front_location].count(None) == 2,\
               Exception('At least one and only one of [x,y,z]_front_location should be specified.')

        self.x_front_location = x_front_location
        self.y_front_location = y_front_location
        self.z_front_location = z_front_location
        self.depth = depth
        self.number_of_cells = number_of_cells

        self.handle_init(kw)

class PICMI_LoadAppliedField(_ClassWithInit):
    """
    The E and B fields read from file are applied to the particles directly. (They are not affected by the field solver.)
    The expected format is the file is OpenPMD with axes (x,y,z) in Cartesian, or (r,z) in Cylindrical geometry.

    Parameters
    ----------
    read_fields_from_path: string
        Path to file with field data

    load_B: bool, default=True
        If False, do not load magnetic field

    load_E: bool, default=True
        If False, do not load electric field
    """
    def __init__(self, read_fields_from_path, load_B=True, load_E=True, **kw):
        self.load_B = load_B
        self.load_E = load_E
        self.read_fields_from_path = read_fields_from_path

        self.handle_init(kw)


class PICMI_LoadGriddedField(_ClassWithInit):
    """
    The data read in is used to initialize the E and B fields on the grid at the start of the simulation.
    The expected format is the file is OpenPMD with axes (x,y,z) in Cartesian, or (r,z) in Cylindrical geometry.

    Parameters
    ----------
    read_fields_from_path: string
        Path to file with field data

    load_B: bool, default=True
        If False, do not load magnetic field

    load_E: bool, default=True
        If False, do not load electric field
    """
    def __init__(self, read_fields_from_path, load_B=True, load_E=True, **kw):
        self.load_B = load_B
        self.load_E = load_E
        self.read_fields_from_path = read_fields_from_path

        self.handle_init(kw)



"""base code for the PICMI standard
"""
import inspect
import warnings

codename = None

# --- The list of supported codes is needed to allow checking for bad arguments.
supported_codes = ['warp', 'warpx', 'fbpic']

def register_codename(_codename):
    """This must be called by the implementing code, passing in the code name"""
    global codename
    codename = _codename

# --- This needs to be set by the implementing package (by calling register_constants).
# --- It allows constants to be used within the picmi interface, with the constants
# --- defined in the implementation.
_implementation_constants = None

def register_constants(implementation_constants):
    """This must be called by the implementing code, passing in the constans object

    Parameters
    ----------
    implementation_constants: python object
        The object must have as attributes the physical constants
    """
    global _implementation_constants
    _implementation_constants = implementation_constants

def _get_constants():
    return _implementation_constants


class _DocumentedMetaClass(type):
    """This is used as a metaclass that combines the __doc__ of the picmistandard base and of the implementation"""
    def __new__(cls, name, bases, attrs):
        # "if bases" skips this for the _ClassWithInit (which has no bases)
        # "if bases[0].__doc__ is not None" skips this for the picmistandard classes since their bases[0] (i.e. _ClassWithInit)
        # has no __doc__.
        if bases and bases[0].__doc__ is not None:
            implementation_doc = attrs.get('__doc__', '')
            if implementation_doc:
                # The format of the added string is intentional.
                # The double return "\n\n" is needed to start a new section in the documentation.
                # Then the four spaces matches the standard level of indentation for doc strings
                # (assuming PEP8 formatting).
                # The final return "\n" assumes that the implementation doc string begins with a return,
                # i.e. a line with only three quotes, """.
                attrs['__doc__'] = bases[0].__doc__ + """\n\n    Implementation specific documentation\n""" + implementation_doc
            else:
                attrs['__doc__'] = bases[0].__doc__
        return super(_DocumentedMetaClass, cls).__new__(cls, name, bases, attrs)


class _ClassWithInit(metaclass=_DocumentedMetaClass):
    def handle_init(self, kw):
        # --- Grab all keywords for the current code.
        # --- Arguments for other supported codes are ignored.
        # --- If there is anything left over, it is an error.
        codekw = {}
        for k,v in kw.copy().items():
            code = k.split('_')[0]
            if code == codename:
                codekw[k] = v
                kw.pop(k)
            elif code in supported_codes:
                kw.pop(k)

        if kw:
            raise TypeError('Unexpected keyword argument: "%s"'%list(kw))

        # --- It is expected that init strips accepted keywords from codekw.
        self.init(codekw)

        if codekw:
            raise TypeError("Unexpected keyword argument for %s: '%s'"%(codename, list(codekw)))

    def init(self, kw):
        # --- The implementation of this routine should use kw.pop() to retrieve input arguments from kw.
        # --- This allows testing for any unused arguments and raising an error if found.
        pass

    def _check_unsupported_argument(self, arg_name, message=None, raise_error=False):
        """Raise a warning or exception if an unsupported argument was specified by the user

        Parameters
        ----------
        arg_name: string
            The name of the unsupported argument

        message: string
            Information to include in the warning/error message

        raise_error: bool
            If False (the default), raise a warning. If true, raise an exception
            (which interrupts the code).

        Implementation note: This should be called in the "init" method of the
        implementing class for each unsupported argument. For example, for the
        'density_scale' argument of Species:

            self._check_unsupported_argument(
                'density_scale',
                message='My code can not handle a density_scale')

        """

        # This compares the value of the parameter with the dault value in the __init__ method.
        # If they differ, this means that the user supplied a value, so a warning or error is raised.
        signature = inspect.signature(self.__init__)
        default_value = signature.parameters[arg_name].default
        if not (getattr(self, arg_name) == default_value):
            full_message = f'{self.__name__}: For argument {arg_name} is not supported.'
            if message is not None:
                full_message += f' {message}'
            if raise_error:
                raise Exception(full_message)
            else:
                warnings.warn(full_message)

    def _unsupported_value(self, arg_name, message='', raise_error=True):
        """Raise a warning or exception for argument with an unsupported value.

        Parameters
        ----------
        arg_name: string
            The name of the argument with an unsupported value

        message: string
            Information to include in the warning/error message

        raise_error: bool
            If False (the default), raise a warning. If true, raise an exception
            (which interrupts the code).

        Implementation note: This should be called when the implementing code handles
        the input arguments. For example, for 'method' in Species:

            if self.method not in ['Boris', 'Li']:
                self._unsupported_value(
                    'method',
                    message='My code only supports Boris and Li')

        """
        full_message = f'{self.__name__}: For argument {arg_name}, the value {getattr(self, arg_name)} is not supported.'
        if message is not None:
            full_message += f' {message}'
        if raise_error:
            raise Exception(full_message)
        else:
            warnings.warn(full_message)

    def _check_deprecated_argument(self, arg_name, message=None, raise_error=False):
        """Raise a warning or exception if a deprecated argument was specified by the user

        Parameters
        ----------
        arg_name: string
            The name of the deprecated argument

        message: string
            Information to include in the warning/error message

        raise_error: bool
            If False (the default), raise a warning. If true, raise an exception
            (which interrupts the code).

        Implementation note: This should be called within PICMI in the "__init__" method of classes
        for each deprecated argument. This assumes that the argument is still included in the
        argument list of __init__ as a transition until it is removed.
        For example, if the 'density_scale' argument of Species was to be deprecated:

            self._check_deprecated_argument(
                'density_scale',
                message='This argument is no longer needed')

        """

        # This compares the value of the parameter with the dault value in the __init__ method.
        # If they differ, this means that the user supplied a value, so a warning or error is raised.
        signature = inspect.signature(self.__init__)
        default_value = signature.parameters[arg_name].default
        if not (getattr(self, arg_name) == default_value):
            full_message = f'{self.__name__}: For argument {arg_name} is not supported.'
            if message is not None:
                full_message += f' {message}'
            if raise_error:
                raise Exception(full_message)
            else:
                warnings.warn(full_message)




"""Classes following the PICMI standard
These should be the base classes for Python implementation of the PICMI standard
"""
import math
import sys

from .base import _ClassWithInit

# ---------------
# Physics objects
# ---------------

class PICMI_ElectromagneticSolver(_ClassWithInit):
    """
    Electromagnetic field solver

    Parameters
    ----------
    grid: grid instance
        Grid object for the diagnostic

    method: {'Yee', 'CKC', 'Lehe', 'PSTD', 'PSATD', 'GPSTD', 'DS', 'ECT'}
        The advance method use to solve Maxwell's equations. The default method is code dependent.

        - 'Yee': standard solver using the staggered Yee grid (https://doi.org/10.1109/TAP.1966.1138693)

        - 'CKC': solver with the extended Cole-Karkkainen-Cowan stencil with better dispersion properties
          (https://doi.org/10.1103/PhysRevSTAB.16.041303)

        - 'Lehe': CKC-style solver with modified dispersion (https://doi.org/10.1103/PhysRevSTAB.16.021301)

        - 'PSTD': Spectral solver with finite difference in time domain, e.g., Q. H. Liu, Letters 15 (3) (1997) 158–165

        - 'PSATD': Spectral solver with analytic in time domain (https://doi.org/10.1016/j.jcp.2013.03.010)

        - 'DS': Directional Splitting after Yasuhiko Sentoku (https://doi.org/10.1140/epjd/e2014-50162-y)

        - 'ECT': Enlarged Cell Technique solver, allowing internal conductors (https://doi.org/10.1109/APS.2005.1551259)

    stencil_order: vector of integers
        Order of stencil for each axis (-1=infinite)

    cfl: float, optional
        Fraction of the Courant-Friedrich-Lewy criteria [1]

    source_smoother: smoother instance, optional
        Smoother object to apply to the sources

    field_smoother: smoother instance, optional
        Smoother object to apply to the fields

    subcycling: integer, optional
        Level of subcycling for the GPSTD solver

    galilean_velocity: vector of floats, optional
        Velocity of Galilean reference frame [m/s]

    divE_cleaning: bool, optional
        Solver uses div(E) cleaning if True

    divB_cleaning: bool, optional
        Solver uses div(B) cleaning if True

    pml_divE_cleaning: bool, optional
        Solver uses div(E) cleaning in the PML if True

    pml_divB_cleaning: bool, optional
        Solver uses div(B) cleaning in the PML if True
    """

    methods_list = ['Yee', 'CKC', 'Lehe', 'PSTD', 'PSATD', 'GPSTD', 'DS', 'ECT']

    def __init__(self, grid, method=None, stencil_order=None, cfl=None,
                 source_smoother=None, field_smoother=None, subcycling=None,
                 galilean_velocity=None, divE_cleaning=None, divB_cleaning=None,
                 pml_divE_cleaning=None, pml_divB_cleaning=None, **kw):

        assert method is None or method in PICMI_ElectromagneticSolver.methods_list, \
               Exception('method must be one of '+', '.join(PICMI_ElectromagneticSolver.methods_list))

        self.grid = grid
        self.method = method
        self.cfl = cfl
        self.stencil_order = stencil_order
        self.source_smoother = source_smoother
        self.field_smoother = field_smoother
        self.subcycling = subcycling
        self.galilean_velocity = galilean_velocity
        self.divE_cleaning = divE_cleaning
        self.divB_cleaning = divB_cleaning
        self.pml_divE_cleaning = pml_divE_cleaning
        self.pml_divB_cleaning = pml_divB_cleaning

        self.handle_init(kw)


class PICMI_ElectrostaticSolver(_ClassWithInit):
    """
    Electrostatic field solver

    Parameters
    ----------
    grid: grid instance
        Grid object for the diagnostic

    method: string
        One of 'FFT', or 'Multigrid'

    required_precision: float, optional
        Level of precision required for iterative solvers

    maximum_iterations: integer, optional
        Maximum number of iterations for iterative solvers
    """

    methods_list = ['FFT', 'Multigrid']

    def __init__(self, grid, method=None,
                 required_precision=None, maximum_iterations=None, **kw):

        assert method is None or method in PICMI_ElectrostaticSolver.methods_list, \
               Exception('method must be one of '+', '.join(PICMI_ElectrostaticSolver.methods_list))

        self.grid = grid
        self.method = method
        self.required_precision = required_precision
        self.maximum_iterations = maximum_iterations

        self.handle_init(kw)


class PICMI_MagnetostaticSolver(_ClassWithInit):
    """
    Magnetostatic field solver

    Parameters
    ----------
    grid: grid instance
        Grid object for the diagnostic

    method: string
        One of 'FFT', or 'Multigrid'
    """

    methods_list = ['FFT', 'Multigrid']

    def __init__(self, grid, method=None, **kw):

        assert method is None or method in PICMI_MagnetostaticSolver.methods_list, \
               Exception('method must be one of '+', '.join(PICMI_MagnetostaticSolver.methods_list))

        self.grid = grid
        self.method = method

        self.handle_init(kw)


# ------------------
# Numeric Objects
# ------------------


class PICMI_BinomialSmoother(_ClassWithInit):
    """
    Describes a binomial smoother operator (applied to grids)

    Parameters
    ----------
    n_pass: vector of integers
        Number of passes along each axis

    compensation: vector of booleans, optional
        Flags whether to apply comensation along each axis

    stride: vector of integers, optional
        Stride along each axis

    alpha: vector of floats, optional
        Smoothing coefficients along each axis
    """
    def __init__(self, n_pass=None, compensation=None, stride=None, alpha=None, **kw):
        self.n_pass = n_pass
        self.compensation = compensation
        self.stride = stride
        self.alpha = alpha

        self.handle_init(kw)


class PICMI_Cartesian1DGrid(_ClassWithInit):
    """
    One-dimensional Cartesian grid
    Parameters can be specified either as vectors or separately.
    (If both are specified, the vector is used.)

    Parameters
    ----------
    number_of_cells: vector of integers
        Number of cells along each axis (number of nodes is number_of_cells+1)

    lower_bound: vector of floats
        Position of the node at the lower bound [m]

    upper_bound: vector of floats
        Position of the node at the upper bound [m]

    lower_boundary_conditions: vector of strings
        Conditions at lower boundaries, periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    upper_boundary_conditions: vector of strings
        Conditions at upper boundaries, periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    nx: integer
        Number of cells along X (number of nodes=nx+1)

    xmin: float
        Position of first node along X [m]

    xmax: float
        Position of last node along X [m]

    bc_xmin: vector of strings
        Boundary condition at min X: One of periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    bc_xmax: vector of strings
        Boundary condition at max X: One of periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    moving_window_velocity: vector of floats, optional
        Moving frame velocity [m/s]

    refined_regions: list of lists, optional
        List of refined regions, each element being a list of the format [level, lo, hi, refinement_factor],
        with level being the refinement level, with 1 being the first level of refinement, 2 being the second etc,
        lo and hi being vectors of length 2 specifying the extent of the region,
        and refinement_factor defaulting to [2,2] (relative to next lower level)

    lower_bound_particles: vector of floats, optional
        Position of particle lower bound [m]

    upper_bound_particles: vector of floats, optional
        Position of particle upper bound [m]

    xmin_particles: float, optional
        Position of min particle boundary along X [m]

    xmax_particles: float, optional
        Position of max particle boundary along X [m]

    lower_boundary_conditions_particles: vector of strings, optional
        Conditions at lower boundaries for particles, periodic, absorbing, reflect or thermal

    upper_boundary_conditions_particles: vector of strings, optional
        Conditions at upper boundaries for particles, periodic, absorbing, reflect or thermal

    bc_xmin_particles: string, optional
        Boundary condition at min X for particles: One of periodic, absorbing, reflect, thermal

    bc_xmax_particles: string, optional
        Boundary condition at max X for particles: One of periodic, absorbing, reflect, thermal

    guard_cells: vector of integers, optional
        Number of guard cells used along each direction

    pml_cells: vector of integers, optional
        Number of Perfectly Matched Layer (PML) cells along each direction

    References
    ----------
    absorbing_silver_mueller: A local absorbing boundary condition that works best under normal incidence angle.
    Based on the Silver-Mueller Radiation Condition, e.g., in

    * A. K. Belhora and L. Pichon, "Maybe Efficient Absorbing Boundary Conditions for the Finite Element Solution of 3D Scattering Problems," 1995,
      https://doi.org/10.1109/20.376322
    * B Engquist and A. Majdat, "Absorbing boundary conditions for numerical simulation of waves," 1977,
      https://doi.org/10.1073/pnas.74.5.1765
    * R. Lehe, "Electromagnetic wave propagation in Particle-In-Cell codes," 2016,
      US Particle Accelerator School (USPAS) Summer Session, Self-Consistent Simulations of Beam and Plasma Systems
      https://people.nscl.msu.edu/~lund/uspas/scs_2016/lec_adv/A1b_EM_Waves.pdf
    """
    # Note for implementations, as a matter of convenience and flexibility, the user interface allows
    # specifying various quantities using either the individual named attributes (such as nx) or a
    # vector of values (such as number_of_cells). However, internally, only the vectors are saved and
    # the implementation needs to use the those to access the user input.

    number_of_dimensions = 1

    def __init__(self, number_of_cells=None, lower_bound=None, upper_bound=None,
                 lower_boundary_conditions=None, upper_boundary_conditions=None,
                 nx=None, xmin=None, xmax=None, bc_xmin=None, bc_xmax=None,
                 moving_window_velocity=None, refined_regions=[],lower_bound_particles=None, upper_bound_particles=None,
                 xmin_particles=None, xmax_particles=None,
                 lower_boundary_conditions_particles=None, upper_boundary_conditions_particles=None,
                 bc_xmin_particles=None, bc_xmax_particles=None,
                 guard_cells=None, pml_cells=None,
                 **kw):

        # Sanity check and init of input arguments related to grid parameters
        assert (number_of_cells is None) and (nx is not None) or \
               (number_of_cells is not None) and (nx is None), \
                Exception('Either number_of_cells or nx must be specified')
        assert (lower_bound is None) and (xmin is not None) or \
               (lower_bound is not None) and (xmin is None), \
                Exception('Either lower_bound or xmin must be specified')
        assert (upper_bound is None) and (xmax is not None) or \
               (upper_bound is not None) and (xmax is None), \
                Exception('Either upper_bound or xmax must be specified')
        assert (lower_boundary_conditions is None) and (bc_xmin is not None) or \
               (lower_boundary_conditions is not None) and (bc_xmin is None), \
                Exception('Either lower_boundary_conditions or bc_xmin')
        assert (upper_boundary_conditions is None) and (bc_xmax is not None) or \
               (upper_boundary_conditions is not None) and (bc_xmax is None), \
                Exception('Either upper_boundary_conditions or bc_xmax must be specified')

        if number_of_cells is None:
            number_of_cells = [nx]
        if lower_bound is None:
            lower_bound = [xmin]
        if upper_bound is None:
            upper_bound = [xmax]
        if lower_boundary_conditions is None:
            lower_boundary_conditions = [bc_xmin]
        if upper_boundary_conditions is None:
            upper_boundary_conditions = [bc_xmax,]

        # Sanity check and init of input arguments related to particle boundary parameters
        # By default, if not specified, particle boundary values are the same as field boundary values
        # By default, if not specified, particle boundary conditions are the same as field boundary conditions
        if lower_bound_particles is None:
            if (xmin_particles is None):
                lower_bound_particles = lower_bound
            else:
                lower_bound_particles = [xmin_particles]
        if upper_bound_particles is None:
            if (xmax_particles is None):
                upper_bound_particles = upper_bound
            else:
                upper_bound_particles=[xmax_particles]

        if lower_boundary_conditions_particles is None:
            if (bc_xmin_particles is None):
                lower_boundary_conditions_particles = lower_boundary_conditions
            else:
                lower_boundary_conditions_particles = [bc_xmin_particles]
        if upper_boundary_conditions_particles is None:
            if (bc_xmax_particles is None):
                upper_boundary_conditions_particles = upper_boundary_conditions
            else:
                upper_boundary_conditions_particles = [bc_xmax_particles]

        # Sanity check on dimensionality of vector quantities
        assert len(number_of_cells) == 1, Exception('Wrong number of cells specified')
        assert len(lower_bound) == 1, Exception('Wrong number of lower bounds specified')
        assert len(upper_bound) == 1, Exception('Wrong number of upper bounds specified')
        assert len(lower_boundary_conditions) == 1, Exception('Wrong number of lower boundary conditions specified')
        assert len(upper_boundary_conditions) == 1, Exception('Wrong number of upper boundary conditions specified')
        assert len(lower_bound_particles) == 1, Exception('Wrong number of particle lower bounds specified')
        assert len(upper_bound_particles) == 1, Exception('Wrong number of particle upper bounds specified')
        assert len(lower_boundary_conditions_particles) == 1, Exception('Wrong number of lower particle boundary conditions specified')
        assert len(upper_boundary_conditions_particles) == 1, Exception('Wrong number of upper particle boundary conditions specified')

        self.number_of_cells = number_of_cells
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.lower_boundary_conditions = lower_boundary_conditions
        self.upper_boundary_conditions = upper_boundary_conditions
        self.lower_bound_particles = lower_bound_particles
        self.upper_bound_particles = upper_bound_particles
        self.lower_boundary_conditions_particles = lower_boundary_conditions_particles
        self.upper_boundary_conditions_particles = upper_boundary_conditions_particles
        self.guard_cells = guard_cells
        self.pml_cells = pml_cells

        self.moving_window_velocity = moving_window_velocity

        self.refined_regions = refined_regions

        for region in self.refined_regions:
            if len(region) == 3:
                region.append([2])
            assert len(region[1]) == 1, Exception('The lo extent of the refined region must be a vector of length 2')
            assert len(region[2]) == 1, Exception('The hi extent of the refined region must be a vector of length 2')
            assert len(region[3]) == 1, Exception('The refinement factor of the refined region must be a vector of length 2')

        self.handle_init(kw)

    def add_refined_region(self, level, lo, hi, refinement_factor=[2]):
        """Add a refined region.
        level: the refinement level, with 1 being the first level of refinement, 2 being the second etc.
        lo, hi: vectors of length 2 specifying the extent of the region
        refinement_factor: defaulting to [2,2] (relative to next lower level)
        """
        self.refined_regions.append([level, lo, hi, refinement_factor])


class PICMI_CylindricalGrid(_ClassWithInit):
    """
    Axisymmetric, cylindrical grid
    Parameters can be specified either as vectors or separately.
    (If both are specified, the vector is used.)

    Parameters
    ----------
    number_of_cells: vector of integers
        Number of cells along each axis (number of nodes is number_of_cells+1)

    lower_bound: vector of floats
        Position of the node at the lower bound [m]

    upper_bound: vector of floats
        Position of the node at the upper bound [m]

    lower_boundary_conditions: vector of strings
        Conditions at lower boundaries, periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    upper_boundary_conditions: vector of strings
        Conditions at upper boundaries, periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    nr: integer
        Number of cells along R (number of nodes=nr+1)

    nz: integer
        Number of cells along Z (number of nodes=nz+1)

    n_azimuthal_modes: integer
        Number of azimuthal modes

    rmin: float
        Position of first node along R [m]

    rmax: float
        Position of last node along R [m]

    zmin: float
        Position of first node along Z [m]

    zmax: float
        Position of last node along Z [m]

    bc_rmin: vector of strings
        Boundary condition at min R: One of open, dirichlet, absorbing_silver_mueller, or neumann

    bc_rmax: vector of strings
        Boundary condition at max R: One of open, dirichlet, absorbing_silver_mueller, or neumann

    bc_zmin: vector of strings
        Boundary condition at min Z: One of periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    bc_zmax: vector of strings
        Boundary condition at max Z: One of periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    moving_window_velocity: vector of floats, optional
        Moving frame velocity [m/s]

    refined_regions: list of lists, optional
        List of refined regions, each element being a list of the format [level, lo, hi, refinement_factor],
        with level being the refinement level, with 1 being the first level of refinement, 2 being the second etc,
        lo and hi being vectors of length 2 specifying the extent of the region,
        and refinement_factor defaulting to [2,2] (relative to next lower level)

    lower_bound_particles: vector of floats, optional
        Position of particle lower bound [m]

    upper_bound_particles: vector of floats, optional
        Position of particle upper bound [m]

    rmin_particles: float, optional
        Position of min particle boundary along R [m]

    rmax_particles: float, optional
        Position of max particle boundary along R [m]

    zmin_particles: float, optional
        Position of min particle boundary along Z [m]

    zmax_particles: float, optional
        Position of max particle boundary along Z [m]

    lower_boundary_conditions_particles: vector of strings, optional
        Conditions at lower boundaries for particles, periodic, absorbing, reflect or thermal

    upper_boundary_conditions_particles: vector of strings, optional
        Conditions at upper boundaries for particles, periodic, absorbing, reflect or thermal

    bc_rmin_particles: string, optional
        Boundary condition at min R for particles: One of periodic, absorbing, reflect, thermal

    bc_rmax_particles: string, optional
        Boundary condition at max R for particles: One of periodic, absorbing, reflect, thermal

    bc_zmin_particles: string, optional
        Boundary condition at min Z for particles: One of periodic, absorbing, reflect, thermal

    bc_zmax_particles: string, optional
        Boundary condition at max Z for particles: One of periodic, absorbing, reflect, thermal

    guard_cells: vector of integers, optional
        Number of guard cells used along each direction

    pml_cells: vector of integers, optional
        Number of Perfectly Matched Layer (PML) cells along each direction

    References
    ----------
    absorbing_silver_mueller: A local absorbing boundary condition that works best under normal incidence angle.
    Based on the Silver-Mueller Radiation Condition, e.g., in

    * A. K. Belhora and L. Pichon, "Maybe Efficient Absorbing Boundary Conditions for the Finite Element Solution of 3D Scattering Problems," 1995,
      https://doi.org/10.1109/20.376322
    * B Engquist and A. Majdat, "Absorbing boundary conditions for numerical simulation of waves," 1977,
      https://doi.org/10.1073/pnas.74.5.1765
    * R. Lehe, "Electromagnetic wave propagation in Particle-In-Cell codes," 2016,
      US Particle Accelerator School (USPAS) Summer Session, Self-Consistent Simulations of Beam and Plasma Systems
      https://people.nscl.msu.edu/~lund/uspas/scs_2016/lec_adv/A1b_EM_Waves.pdf
    """
    # Note for implementations, as a matter of convenience and flexibility, the user interface allows
    # specifying various quantities using either the individual named attributes (such as nr and nz) or a
    # vector of values (such as number_of_cells). However, internally, only the vectors are saved and
    # the implementation needs to use the those to access the user input.

    number_of_dimensions = 2

    def __init__(self, number_of_cells=None, lower_bound=None, upper_bound=None,
                 lower_boundary_conditions=None, upper_boundary_conditions=None,
                 nr=None, nz=None, n_azimuthal_modes=None,
                 rmin=None, rmax=None, zmin=None, zmax=None,
                 bc_rmin=None, bc_rmax=None, bc_zmin=None, bc_zmax=None,
                 moving_window_velocity=None, refined_regions=[],
                 lower_bound_particles=None, upper_bound_particles=None,
                 rmin_particles=None, rmax_particles=None, zmin_particles=None, zmax_particles=None,
                 lower_boundary_conditions_particles=None, upper_boundary_conditions_particles=None,
                 bc_rmin_particles=None, bc_rmax_particles=None, bc_zmin_particles=None, bc_zmax_particles=None,
                 guard_cells=None, pml_cells=None, **kw):

        # Sanity check and init of input arguments related to grid parameters
        assert (number_of_cells is None) and (nr is not None and nz is not None) or \
               (number_of_cells is not None) and (nr is None and nz is None), \
                Exception('Either number_of_cells or nr and nz must be specified')
        assert (lower_bound is None) and (rmin is not None and zmin is not None) or \
               (lower_bound is not None) and (rmin is None and zmin is None), \
                Exception('Either lower_bound or rmin and zmin must be specified')
        assert (upper_bound is None) and (rmax is not None and zmax is not None) or \
               (upper_bound is not None) and (rmax is None and zmax is None), \
                Exception('Either upper_bound or rmax and zmax must be specified')
        # --Allow bc_rmin to be None since it will usually be the axis.
        assert (lower_boundary_conditions is None) and (bc_zmin is not None) or \
               (lower_boundary_conditions is not None) and (bc_rmin is None and bc_zmin is None), \
                Exception('Either lower_boundary_conditions or bc_rmin and bc_zmin must be specified')
        assert (upper_boundary_conditions is None) and (bc_rmax is not None and bc_zmax is not None) or \
               (upper_boundary_conditions is not None) and (bc_rmax is None and bc_zmax is None), \
                Exception('Either upper_boundary_conditions or bc_rmax and bc_zmax must be specified')

        if number_of_cells is None:
            number_of_cells = [nr, nz]
        if lower_bound is None:
            lower_bound = [rmin, zmin]
        if upper_bound is None:
            upper_bound = [rmax, zmax]
        if lower_boundary_conditions is None:
            lower_boundary_conditions = [bc_rmin, bc_zmin]
        if upper_boundary_conditions is None:
            upper_boundary_conditions = [bc_rmax, bc_zmax]

        # Sanity check and init of input arguments related to particle boundary parameters
        # By default, if not specified, particle boundary values are the same as field boundary values
        # By default, if not specified, particle boundary conditions are the same as field boundary conditions
        if lower_bound_particles is None:
            if (rmin_particles is None) and (zmin_particles is None):
                lower_bound_particles = lower_bound
            else:
                lower_bound_particles = [rmin_particles, zmin_particles]
        if upper_bound_particles is None:
            if (rmax_particles is None) and (zmax_particles is None):
                upper_bound_particles = upper_bound
            else:
                upper_bound_particles=[rmax_particles, zmax_particles]

        if lower_boundary_conditions_particles is None:
            if (bc_rmin_particles is None) and (bc_zmin_particles is None):
                lower_boundary_conditions_particles = lower_boundary_conditions
            else:
                lower_boundary_conditions_particles = [bc_rmin_particles, bc_zmin_particles]
        if upper_boundary_conditions_particles is None:
            if (bc_rmax_particles is None) and (bc_zmax_particles is None):
                upper_boundary_conditions_particles = upper_boundary_conditions
            else:
                upper_boundary_conditions_particles = [bc_rmax_particles, bc_zmax_particles]

        # Sanity check on dimensionality of vector quantities
        assert len(number_of_cells) == 2, Exception('Wrong number of cells specified')
        assert len(lower_bound) == 2, Exception('Wrong number of lower bounds specified')
        assert len(upper_bound) == 2, Exception('Wrong number of upper bounds specified')
        assert len(lower_boundary_conditions) == 2, Exception('Wrong number of lower boundary conditions specified')
        assert len(upper_boundary_conditions) == 2, Exception('Wrong number of upper boundary conditions specified')

        self.number_of_cells = number_of_cells

        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.lower_boundary_conditions = lower_boundary_conditions
        self.upper_boundary_conditions = upper_boundary_conditions
        self.lower_bound_particles = lower_bound_particles
        self.upper_bound_particles = upper_bound_particles
        self.lower_boundary_conditions_particles = lower_boundary_conditions_particles
        self.upper_boundary_conditions_particles = upper_boundary_conditions_particles
        self.guard_cells = guard_cells
        self.pml_cells = pml_cells

        self.n_azimuthal_modes = n_azimuthal_modes

        self.moving_window_velocity = moving_window_velocity

        self.refined_regions = refined_regions
        for region in self.refined_regions:
            if len(region) == 3:
                region.append([2,2])
            assert len(region[1]) == 2, Exception('The lo extent of the refined region must be a vector of length 2')
            assert len(region[2]) == 2, Exception('The hi extent of the refined region must be a vector of length 2')
            assert len(region[3]) == 2, Exception('The refinement factor of the refined region must be a vector of length 2')

        self.handle_init(kw)

    def add_refined_region(self, level, lo, hi, refinement_factor=[2,2]):
        """Add a refined region.
        level: the refinement level, with 1 being the first level of refinement, 2 being the second etc.
        lo, hi: vectors of length 2 specifying the extent of the region
        refinement_factor: defaulting to [2,2] (relative to next lower level)
        """
        self.refined_regions.append([level, lo, hi, refinement_factor])


class PICMI_Cartesian2DGrid(_ClassWithInit):
    """
    Two dimensional Cartesian grid
    Parameters can be specified either as vectors or separately.
    (If both are specified, the vector is used.)

    Parameters
    ----------
    number_of_cells: vector of integers
        Number of cells along each axis (number of nodes is number_of_cells+1)

    lower_bound: vector of floats
        Position of the node at the lower bound [m]

    upper_bound: vector of floats
        Position of the node at the upper bound [m]

    lower_boundary_conditions: vector of strings
        Conditions at lower boundaries, periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    upper_boundary_conditions: vector of strings
        Conditions at upper boundaries, periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    nx: integer
        Number of cells along X (number of nodes=nx+1)

    ny: integer
        Number of cells along Y (number of nodes=ny+1)

    xmin: float
        Position of first node along X [m]

    xmax: float
        Position of last node along X [m]

    ymin: float
        Position of first node along Y [m]

    ymax: float
        Position of last node along Y [m]

    bc_xmin: vector of strings
        Boundary condition at min X: One of periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    bc_xmax: vector of strings
        Boundary condition at max X: One of periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    bc_ymin: vector of strings
        Boundary condition at min Y: One of periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    bc_ymax: vector of strings
        Boundary condition at max Y: One of periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    moving_window_velocity: vector of floats, optional
        Moving frame velocity [m/s]

    refined_regions: list of lists, optional
        List of refined regions, each element being a list of the format [level, lo, hi, refinement_factor],
        with level being the refinement level, with 1 being the first level of refinement, 2 being the second etc,
        lo and hi being vectors of length 2 specifying the extent of the region,
        and refinement_factor defaulting to [2,2] (relative to next lower level)

    lower_bound_particles: vector of floats, optional
        Position of particle lower bound [m]

    upper_bound_particles: vector of floats, optional
        Position of particle upper bound [m]

    xmin_particles: float, optional
        Position of min particle boundary along X [m]

    xmax_particles: float, optional
        Position of max particle boundary along X [m]

    ymin_particles: float, optional
        Position of min particle boundary along Y [m]

    ymax_particles: float, optional
        Position of max particle boundary along Y [m]

    lower_boundary_conditions_particles: vector of strings, optional
        Conditions at lower boundaries for particles, periodic, absorbing, reflect or thermal

    upper_boundary_conditions_particles: vector of strings, optional
        Conditions at upper boundaries for particles, periodic, absorbing, reflect or thermal

    bc_xmin_particles: string, optional
        Boundary condition at min X for particles: One of periodic, absorbing, reflect, thermal

    bc_xmax_particles: string, optional
        Boundary condition at max X for particles: One of periodic, absorbing, reflect, thermal

    bc_ymin_particles: string, optional
        Boundary condition at min Y for particles: One of periodic, absorbing, reflect, thermal

    bc_ymax_particles: string, optional
        Boundary condition at max Y for particles: One of periodic, absorbing, reflect, thermal

    guard_cells: vector of integers, optional
        Number of guard cells used along each direction

    pml_cells: vector of integers, optional
        Number of Perfectly Matched Layer (PML) cells along each direction

    References
    ----------
    absorbing_silver_mueller: A local absorbing boundary condition that works best under normal incidence angle.
    Based on the Silver-Mueller Radiation Condition, e.g., in

    * A. K. Belhora and L. Pichon, "Maybe Efficient Absorbing Boundary Conditions for the Finite Element Solution of 3D Scattering Problems," 1995,
      https://doi.org/10.1109/20.376322
    * B Engquist and A. Majdat, "Absorbing boundary conditions for numerical simulation of waves," 1977,
      https://doi.org/10.1073/pnas.74.5.1765
    * R. Lehe, "Electromagnetic wave propagation in Particle-In-Cell codes," 2016,
      US Particle Accelerator School (USPAS) Summer Session, Self-Consistent Simulations of Beam and Plasma Systems
      https://people.nscl.msu.edu/~lund/uspas/scs_2016/lec_adv/A1b_EM_Waves.pdf
    """
    # Note for implementations, as a matter of convenience and flexibility, the user interface allows
    # specifying various quantities using either the individual named attributes (such as nx and ny) or a
    # vector of values (such as number_of_cells). However, internally, only the vectors are saved and
    # the implementation needs to use the those to access the user input.

    number_of_dimensions = 2

    def __init__(self, number_of_cells=None, lower_bound=None, upper_bound=None,
                 lower_boundary_conditions=None, upper_boundary_conditions=None,
                 nx=None, ny=None,
                 xmin=None, xmax=None, ymin=None, ymax=None,
                 bc_xmin=None, bc_xmax=None, bc_ymin=None, bc_ymax=None,
                 moving_window_velocity=None, refined_regions=[],lower_bound_particles=None, upper_bound_particles=None,
                 xmin_particles=None, xmax_particles=None, ymin_particles=None, ymax_particles=None,
                 lower_boundary_conditions_particles=None, upper_boundary_conditions_particles=None,
                 bc_xmin_particles=None, bc_xmax_particles=None, bc_ymin_particles=None, bc_ymax_particles=None,
                 guard_cells=None, pml_cells=None,
                 **kw):

        # Sanity check and init of input arguments related to grid parameters
        assert (number_of_cells is None) and (nx is not None and ny is not None) or \
               (number_of_cells is not None) and (nx is None and ny is None), \
                Exception('Either number_of_cells or nx and ny must be specified')
        assert (lower_bound is None) and (xmin is not None and ymin is not None) or \
               (lower_bound is not None) and (xmin is None and ymin is None), \
                Exception('Either lower_bound or xmin and ymin must be specified')
        assert (upper_bound is None) and (xmax is not None and ymax is not None) or \
               (upper_bound is not None) and (xmax is None and ymax is None), \
                Exception('Either upper_bound or xmax and ymax must be specified')
        assert (lower_boundary_conditions is None) and (bc_xmin is not None and bc_ymin is not None) or \
               (lower_boundary_conditions is not None) and (bc_xmin is None and bc_ymin is None), \
                Exception('Either lower_boundary_conditions or bc_xmin and bc_ymin must be specified')
        assert (upper_boundary_conditions is None) and (bc_xmax is not None and bc_ymax is not None) or \
               (upper_boundary_conditions is not None) and (bc_xmax is None and bc_ymax is None), \
                Exception('Either upper_boundary_conditions or bc_xmax and bc_ymax must be specified')

        if number_of_cells is None:
            number_of_cells = [nx, ny]
        if lower_bound is None:
            lower_bound = [xmin, ymin]
        if upper_bound is None:
            upper_bound = [xmax, ymax]
        if lower_boundary_conditions is None:
            lower_boundary_conditions = [bc_xmin, bc_ymin]
        if upper_boundary_conditions is None:
            upper_boundary_conditions = [bc_xmax, bc_ymax]

        # Sanity check and init of input arguments related to particle boundary parameters
        # By default, if not specified, particle boundary values are the same as field boundary values
        # By default, if not specified, particle boundary conditions are the same as field boundary conditions
        if lower_bound_particles is None:
            if (xmin_particles is None) and (ymin_particles is None):
                lower_bound_particles = lower_bound
            else:
                lower_bound_particles = [xmin_particles, ymin_particles]
        if upper_bound_particles is None:
            if (xmax_particles is None) and (ymax_particles is None):
                upper_bound_particles = upper_bound
            else:
                upper_bound_particles=[xmax_particles, ymax_particles]

        if lower_boundary_conditions_particles is None:
            if (bc_xmin_particles is None) and (bc_ymin_particles is None):
                lower_boundary_conditions_particles = lower_boundary_conditions
            else:
                lower_boundary_conditions_particles = [bc_xmin_particles, bc_ymin_particles]
        if upper_boundary_conditions_particles is None:
            if (bc_xmax_particles is None) and (bc_ymax_particles is None):
                upper_boundary_conditions_particles = upper_boundary_conditions
            else:
                upper_boundary_conditions_particles = [bc_xmax_particles, bc_ymax_particles]

        # Sanity check on dimensionality of vector quantities
        assert len(number_of_cells) == 2, Exception('Wrong number of cells specified')
        assert len(lower_bound) == 2, Exception('Wrong number of lower bounds specified')
        assert len(upper_bound) == 2, Exception('Wrong number of upper bounds specified')
        assert len(lower_boundary_conditions) == 2, Exception('Wrong number of lower boundary conditions specified')
        assert len(upper_boundary_conditions) == 2, Exception('Wrong number of upper boundary conditions specified')
        assert len(lower_bound_particles) == 2, Exception('Wrong number of particle lower bounds specified')
        assert len(upper_bound_particles) == 2, Exception('Wrong number of particle upper bounds specified')
        assert len(lower_boundary_conditions_particles) == 2, Exception('Wrong number of lower particle boundary conditions specified')
        assert len(upper_boundary_conditions_particles) == 2, Exception('Wrong number of upper particle boundary conditions specified')

        self.number_of_cells = number_of_cells
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.lower_boundary_conditions = lower_boundary_conditions
        self.upper_boundary_conditions = upper_boundary_conditions
        self.lower_bound_particles = lower_bound_particles
        self.upper_bound_particles = upper_bound_particles
        self.lower_boundary_conditions_particles = lower_boundary_conditions_particles
        self.upper_boundary_conditions_particles = upper_boundary_conditions_particles
        self.guard_cells = guard_cells
        self.pml_cells = pml_cells

        self.moving_window_velocity = moving_window_velocity

        self.refined_regions = refined_regions

        for region in self.refined_regions:
            if len(region) == 3:
                region.append([2,2])
            assert len(region[1]) == 2, Exception('The lo extent of the refined region must be a vector of length 2')
            assert len(region[2]) == 2, Exception('The hi extent of the refined region must be a vector of length 2')
            assert len(region[3]) == 2, Exception('The refinement factor of the refined region must be a vector of length 2')

        self.handle_init(kw)

    def add_refined_region(self, level, lo, hi, refinement_factor=[2,2]):
        """Add a refined region.
        level: the refinement level, with 1 being the first level of refinement, 2 being the second etc.
        lo, hi: vectors of length 2 specifying the extent of the region
        refinement_factor: defaulting to [2,2] (relative to next lower level)
        """
        self.refined_regions.append([level, lo, hi, refinement_factor])


class PICMI_Cartesian3DGrid(_ClassWithInit):
    """
    Three dimensional Cartesian grid
    Parameters can be specified either as vectors or separately.
    (If both are specified, the vector is used.)

    Parameters
    ----------
    number_of_cells: vector of integers
        Number of cells along each axis (number of nodes is number_of_cells+1)

    lower_bound: vector of floats
        Position of the node at the lower bound [m]

    upper_bound: vector of floats
        Position of the node at the upper bound [m]

    lower_boundary_conditions: vector of strings
        Conditions at lower boundaries, periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    upper_boundary_conditions: vector of strings
        Conditions at upper boundaries, periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    nx: integer
        Number of cells along X (number of nodes=nx+1)

    ny: integer
        Number of cells along Y (number of nodes=ny+1)

    nz: integer
        Number of cells along Z (number of nodes=nz+1)

    xmin: float
        Position of first node along X [m]

    xmax: float
        Position of last node along X [m]

    ymin: float
        Position of first node along Y [m]

    ymax: float
        Position of last node along Y [m]

    zmin: float
        Position of first node along Z [m]

    zmax: float
        Position of last node along Z [m]

    bc_xmin: string
        Boundary condition at min X: One of periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    bc_xmax: string
        Boundary condition at max X: One of periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    bc_ymin: string
        Boundary condition at min Y: One of periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    bc_ymax: string
        Boundary condition at max Y: One of periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    bc_zmin: string
        Boundary condition at min Z: One of periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    bc_zmax: string
        Boundary condition at max Z: One of periodic, open, dirichlet, absorbing_silver_mueller, or neumann

    moving_window_velocity: vector of floats, optional
        Moving frame velocity [m/s]

    refined_regions: list of lists, optional
        List of refined regions, each element being a list of the format [level, lo, hi, refinement_factor],
        with level being the refinement level, with 1 being the first level of refinement, 2 being the second etc,
        lo and hi being vectors of length 3 specifying the extent of the region,
        and refinement_factor defaulting to [2,2,2] (relative to next lower level)

    lower_bound_particles: vector of floats, optional
        Position of particle lower bound [m]

    upper_bound_particles: vector of floats, optional
        Position of particle upper bound [m]

    xmin_particles: float, optional
        Position of min particle boundary along X [m]

    xmax_particles: float, optional
        Position of max particle boundary along X [m]

    ymin_particles: float, optional
        Position of min particle boundary along Y [m]

    ymax_particles: float, optional
        Position of max particle boundary along Y [m]

    zmin_particles float, optional
        Position of min particle boundary along Z [m]

    zmax_particles: float, optional
        Position of max particle boundary along Z [m]

    lower_boundary_conditions_particles: vector of strings, optional
        Conditions at lower boundaries for particles, periodic, absorbing, reflect or thermal

    upper_boundary_conditions_particles: vector of strings, optional
        Conditions at upper boundaries for particles, periodic, absorbing, reflect or thermal

    bc_xmin_particles: string, optional
        Boundary condition at min X for particles: One of periodic, absorbing, reflect, thermal

    bc_xmax_particles: string, optional
        Boundary condition at max X for particles: One of periodic, absorbing, reflect, thermal

    bc_ymin_particles: string, optional
        Boundary condition at min Y for particles: One of periodic, absorbing, reflect, thermal

    bc_ymax_particles: string, optional
        Boundary condition at max Y for particles: One of periodic, absorbing, reflect, thermal

    bc_zmin_particles: string, optional
        Boundary condition at min Z for particles: One of periodic, absorbing, reflect, thermal

    bc_zmax_particles: string, optional
        Boundary condition at max Z for particles: One of periodic, absorbing, reflect, thermal

    guard_cells: vector of integers, optional
        Number of guard cells used along each direction

    pml_cells: vector of integers, optional
        Number of Perfectly Matched Layer (PML) cells along each direction

    References
    ----------
    absorbing_silver_mueller: A local absorbing boundary condition that works best under normal incidence angle.
    Based on the Silver-Mueller Radiation Condition, e.g., in

    * A. K. Belhora and L. Pichon, "Maybe Efficient Absorbing Boundary Conditions for the Finite Element Solution of 3D Scattering Problems," 1995,
      https://doi.org/10.1109/20.376322
    * B Engquist and A. Majdat, "Absorbing boundary conditions for numerical simulation of waves," 1977,
      https://doi.org/10.1073/pnas.74.5.1765
    * R. Lehe, "Electromagnetic wave propagation in Particle-In-Cell codes," 2016,
      US Particle Accelerator School (USPAS) Summer Session, Self-Consistent Simulations of Beam and Plasma Systems
      https://people.nscl.msu.edu/~lund/uspas/scs_2016/lec_adv/A1b_EM_Waves.pdf
    """
    # Note for implementations, as a matter of convenience and flexibility, the user interface allows
    # specifying various quantities using either the individual named attributes (such as nx, ny, and nz) or a
    # vector of values (such as number_of_cells). However, internally, only the vectors are saved and
    # the implementation needs to use the those to access the user input.

    number_of_dimensions = 3

    def __init__(self, number_of_cells=None, lower_bound=None, upper_bound=None,
                 lower_boundary_conditions=None, upper_boundary_conditions=None,
                 nx=None, ny=None, nz=None,
                 xmin=None, xmax=None, ymin=None, ymax=None, zmin=None, zmax=None,
                 bc_xmin=None, bc_xmax=None, bc_ymin=None, bc_ymax=None, bc_zmin=None, bc_zmax=None,
                 moving_window_velocity=None, refined_regions=[], lower_bound_particles=None, upper_bound_particles=None,
                 xmin_particles=None, xmax_particles=None, ymin_particles=None, ymax_particles=None, zmin_particles=None, zmax_particles=None,
                 lower_boundary_conditions_particles=None, upper_boundary_conditions_particles=None,
                 bc_xmin_particles=None, bc_xmax_particles=None, bc_ymin_particles=None, bc_ymax_particles=None,
                 bc_zmin_particles=None, bc_zmax_particles=None, guard_cells=None, pml_cells=None,
                 **kw):

        # Sanity check and init of input arguments related to grid parameters
        assert (number_of_cells is None) and (nx is not None and ny is not None and nz is not None) or \
               (number_of_cells is not None) and (nx is None and ny is None and nz is None), \
                Exception('Either number_of_cells or nx, ny, and nz must be specified')
        assert (lower_bound is None) and (xmin is not None and ymin is not None and zmin is not None) or \
               (lower_bound is not None) and (xmin is None and ymin is None and zmin is None), \
                Exception('Either lower_bound or xmin, ymin, and zmin must be specified')
        assert (upper_bound is None) and (xmax is not None and ymax is not None and zmax is not None) or \
               (upper_bound is not None) and (xmax is None and ymax is None and zmax is None), \
                Exception('Either upper_bound or xmax, ymax, and zmax must be specified')
        assert (lower_boundary_conditions is None) and (bc_xmin is not None and bc_ymin is not None and bc_zmin is not None) or \
               (lower_boundary_conditions is not None) and (bc_xmin is None and bc_ymin is None and bc_zmin is None), \
                Exception('Either lower_boundary_conditions or bc_xmin, bc_ymin, and bc_zmin must be specified')
        assert (upper_boundary_conditions is None) and (bc_xmax is not None and bc_ymax is not None and bc_zmax is not None) or \
               (upper_boundary_conditions is not None) and (bc_xmax is None and bc_ymax is None and bc_zmax is None), \
                Exception('Either upper_boundary_conditions or bc_xmax, bc_ymax, and bc_zmax must be specified')

        if number_of_cells is None:
            number_of_cells = [nx, ny, nz]
        if lower_bound is None:
            lower_bound = [xmin, ymin, zmin]
        if upper_bound is None:
            upper_bound = [xmax, ymax, zmax]
        if lower_boundary_conditions is None:
            lower_boundary_conditions = [bc_xmin, bc_ymin, bc_zmin]
        if upper_boundary_conditions is None:
            upper_boundary_conditions = [bc_xmax, bc_ymax, bc_zmax]

        # Sanity check and init of input arguments related to particle boundary parameters
        # By default, if not specified, particle boundary values are the same as field boundary values
        # By default, if not specified, particle boundary conditions are the same as field boundary conditions
        if lower_bound_particles is None:
            if (xmin_particles is None) and (ymin_particles is None) and (zmin_particles is None):
                lower_bound_particles = lower_bound
            else:
                lower_bound_particles = [xmin_particles, ymin_particles, zmin_particles]
        if upper_bound_particles is None:
            if (xmax_particles is None) and (ymax_particles is None) and (zmax_particles is None):
                upper_bound_particles = upper_bound
            else:
                upper_bound_particles = [xmax_particles, ymax_particles, zmax_particles]

        if lower_boundary_conditions_particles is None:
            if (bc_xmin_particles is None) and (bc_ymin_particles is None) and (bc_zmin_particles is None):
                lower_boundary_conditions_particles = lower_boundary_conditions
            else:
                lower_boundary_conditions_particles = [bc_xmin_particles, bc_ymin_particles, bc_zmin_particles]
        if upper_boundary_conditions_particles is None:
            if (bc_xmax_particles is None) and (bc_ymax_particles is None) and (bc_zmax_particles is None):
                upper_boundary_conditions_particles = upper_boundary_conditions
            else:
                upper_boundary_conditions_particles = [bc_xmax_particles, bc_ymax_particles, bc_zmax_particles]

        # Sanity check on number of arguments of vector quantities
        assert len(number_of_cells) == 3, Exception('Wrong number of cells specified')
        assert len(lower_bound) == 3, Exception('Wrong number of lower bounds specified')
        assert len(upper_bound) == 3, Exception('Wrong number of upper bounds specified')
        assert len(lower_boundary_conditions) == 3, Exception('Wrong number of lower boundary conditions specified')
        assert len(upper_boundary_conditions) == 3, Exception('Wrong number of upper boundary conditions specified')
        assert len(lower_bound_particles) == 3, Exception('Wrong number of particle lower bounds specified')
        assert len(upper_bound_particles) == 3, Exception('Wrong number of particle upper bounds specified')
        assert len(lower_boundary_conditions_particles) == 3, Exception('Wrong number of particle lower boundary conditions specified')
        assert len(upper_boundary_conditions_particles) == 3, Exception('Wrong number of particle upper boundary conditions specified')

        self.number_of_cells = number_of_cells
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.lower_boundary_conditions = lower_boundary_conditions
        self.upper_boundary_conditions = upper_boundary_conditions
        self.lower_bound_particles = lower_bound_particles
        self.upper_bound_particles = upper_bound_particles
        self.lower_boundary_conditions_particles = lower_boundary_conditions_particles
        self.upper_boundary_conditions_particles = upper_boundary_conditions_particles
        self.guard_cells = guard_cells
        self.pml_cells = pml_cells

        self.moving_window_velocity = moving_window_velocity

        self.refined_regions = refined_regions
        for region in self.refined_regions:
            if len(region) == 3:
                region.append([2,2,2])
            assert len(region[1]) == 3, Exception('The lo extent of the refined region must be a vector of length 3')
            assert len(region[2]) == 3, Exception('The hi extent of the refined region must be a vector of length 3')
            assert len(region[3]) == 3, Exception('The refinement factor of the refined region must be a vector of length 3')

        self.handle_init(kw)

    def add_refined_region(self, level, lo, hi, refinement_factor=[2,2,2]):
        """Add a refined region.

        Parameters
        ----------
        level: integer
            The refinement level, with 1 being the first level of refinement, 2 being the second etc.

        lo, hi: vectors of floats
            Each is a vector of length 3 specifying the extent of the region

        refinement_factor: vector of integers, optional
            Defaulting to [2,2,2] (relative to next lower level)
        """
        self.refined_regions.append([level, lo, hi, refinement_factor])



"""Classes following the PICMI standard
These should be the base classes for Python implementation of the PICMI standard
The classes in the file are all diagnostics related
"""

from .base import _ClassWithInit

# ----------------------------
# Simulation frame diagnostics
# ----------------------------


class PICMI_FieldDiagnostic(_ClassWithInit):
    """
    Defines the electromagnetic field diagnostics in the simulation frame

    Parameters
    ----------
    grid: grid instance
        Grid object for the diagnostic

    period: integer
        Period of time steps that the diagnostic is performed

    data_list: list of strings, optional
        List of quantities to write out. Possible values 'rho', 'E', 'B', 'J', 'Ex' etc.
        Defaults to the output list of the implementing code.

    write_dir: string, optional
        Directory where data is to be written

    step_min: integer, default=0
        Minimum step at which diagnostics could be written

    step_max: integer, default=unbounded
        Maximum step at which diagnostics could be written

    number_of_cells: vector of integers, optional
        Number of cells in each dimension.
        If not given, will be obtained from grid.

    lower_bound: vector of floats, optional
        Lower corner of diagnostics box in each direction.
        If not given, will be obtained from grid.

    upper_bound: vector of floats, optional
        Higher corner of diagnostics box in each direction.
        If not given, will be obtained from grid.

    parallelio: bool, optional
        If set to True, field diagnostics are dumped in parallel

    name: string, optional
        Sets the base name for the diagnostic output files
    """
    def __init__(self, grid, period, data_list=None,
                 write_dir = None,
                 step_min = None,
                 step_max = None,
                 number_of_cells = None,
                 lower_bound = None,
                 upper_bound = None,
                 parallelio = None,
                 name = None,
                 **kw):

        if data_list is not None:
            assert isinstance(data_list, list), 'FieldDiagnostic: data_list must be a list'

        self.grid = grid
        self.period = period
        self.data_list = data_list
        self.write_dir = write_dir
        self.step_min = step_min
        self.step_max = step_max
        self.number_of_cells = number_of_cells
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.parallelio = parallelio
        self.name = name

        self.handle_init(kw)


class PICMI_ElectrostaticFieldDiagnostic(_ClassWithInit):
    """
    Defines the electrostatic field diagnostics in the simulation frame

    Parameters
    ----------
    grid: grid instance
        Grid object for the diagnostic

    period: integer
        Period of time steps that the diagnostic is performed

    data_list: list of strings, optional
        List of quantities to write out. Possible values 'rho', 'E', 'B', 'Ex' etc.
        Defaults to the output list of the implementing code.

    write_dir: string, optional
        Directory where data is to be written

    step_min: integer, default=0
        Minimum step at which diagnostics could be written

    step_max: integer, default=unbounded
        Maximum step at which diagnostics could be written

    number_of_cells: vector of integers, optional
        Number of cells in each dimension.
        If not given, will be obtained from grid.

    lower_bound: vector of floats, optional
        Lower corner of diagnostics box in each direction.
        If not given, will be obtained from grid.

    upper_bound: vector of floats, optional
        Higher corner of diagnostics box in each direction.
        If not given, will be obtained from grid.

    parallelio: bool, optional
        If set to True, field diagnostics are dumped in parallel

    name: string, optional
        Sets the base name for the diagnostic output files
    """
    def __init__(self, grid, period, data_list=None,
                 write_dir = None,
                 step_min = None,
                 step_max = None,
                 number_of_cells = None,
                 lower_bound = None,
                 upper_bound = None,
                 parallelio = None,
                 name = None,
                 **kw):

        if data_list is not None:
            assert isinstance(data_list, list), 'ElectrostaticFieldDiagnostic: data_list must be a list'

        self.grid = grid
        self.period = period
        self.data_list = data_list
        self.write_dir = write_dir
        self.step_min = step_min
        self.step_max = step_max
        self.number_of_cells = number_of_cells
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.parallelio = parallelio
        self.name = name

        self.handle_init(kw)


class PICMI_ParticleDiagnostic(_ClassWithInit) :
    """
    Defines the particle diagnostics in the simulation frame

    Parameters
    ----------
    period: integer
        Period of time steps that the diagnostic is performed

    species: species instance or list of species instances, optional
        Species to write out. If not specified, all species are written.
        Note that the name attribute must be defined for the species.

    data_list: list of strings, optional
        The data to be written out. Possible values 'position', 'momentum', 'weighting'.
        Defaults to the output list of the implementing code.

    write_dir: string, optional
        Directory where data is to be written

    step_min: integer, default=0
        Minimum step at which diagnostics could be written

    step_max: integer, default=unbounded
        Maximum step at which diagnostics could be written

    parallelio: bool, optional
        If set to True, particle diagnostics are dumped in parallel

    name: string, optional
        Sets the base name for the diagnostic output files
    """

    def __init__(self, period, species=None, data_list=None,
                 write_dir = None,
                 step_min = None,
                 step_max = None,
                 parallelio = None,
                 name = None,
                 **kw):

        if data_list is not None:
            assert isinstance(data_list, list), 'ParticleDiagnostic: data_list must be a list'

        self.period = period
        self.species = species
        self.data_list = data_list
        self.write_dir = write_dir
        self.step_min = step_min
        self.step_max = step_max
        self.parallelio = parallelio
        self.name = name

        self.handle_init(kw)


class PICMI_ParticleBoundaryScrapingDiagnostic(_ClassWithInit) :
    """
    Defines the particle diagnostics that are used to collect the particles that are absorbed at the boundaries, throughout the simulation.

    Parameters
    ----------
    period: integer
        Period of time steps that the diagnostic is performed

    species: species instance or list of species instances, optional
        Species to write out. If not specified, all species are written.
        Note that the name attribute must be defined for the species.

    data_list: list of strings, optional
        The data to be written out. Possible values 'position', 'momentum', 'weighting'.
        Defaults to the output list of the implementing code.

    write_dir: string, optional
        Directory where data is to be written

    parallelio: bool, optional
        If set to True, particle diagnostics are dumped in parallel

    name: string, optional
        Sets the base name for the diagnostic output files
    """

    def __init__(self, period, species=None, data_list=None,
                 write_dir = None,
                 parallelio = None,
                 name = None,
                 **kw):

        if data_list is not None:
            assert isinstance(data_list, list), 'ParticleBoundaryScrapingDiagnostic: data_list must be a list'

        self.period = period
        self.species = species
        self.data_list = data_list
        self.write_dir = write_dir
        self.parallelio = parallelio
        self.name = name

        self.handle_init(kw)

# ----------------------------
# Lab frame diagnostics
# ----------------------------


class PICMI_LabFrameFieldDiagnostic(_ClassWithInit):
    """
    Defines the electromagnetic field diagnostics in the lab frame

    Parameters
    ----------
    grid: grid instance
        Grid object for the diagnostic

    num_snapshots: integer
        Number of lab frame snapshots to make

    dt_snapshots: float
        Time between each snapshot in lab frame

    data_list: list of strings, optional
        List of quantities to write out. Possible values 'rho', 'E', 'B', 'J', 'Ex' etc.
        Defaults to the output list of the implementing code.

    z_subsampling: integer, default=1
        A factor which is applied on the resolution of the lab frame reconstruction

    time_start: float, default=0
        Time for the first snapshot in lab frame

    write_dir: string, optional
        Directory where data is to be written

    parallelio: bool, optional
        If set to True, field diagnostics are dumped in parallel

    name: string, optional
        Sets the base name for the diagnostic output files
    """
    def __init__(self, grid, num_snapshots, dt_snapshots, data_list=None,
                 z_subsampling = 1, time_start = 0.,
                 write_dir = None,
                 parallelio = None,
                 name = None,
                 **kw):

        if data_list is not None:
            assert isinstance(data_list, list), 'LabFrameFieldDiagnostic: data_list must be a list'

        self.grid = grid
        self.num_snapshots = num_snapshots
        self.dt_snapshots = dt_snapshots
        self.z_subsampling = z_subsampling
        self.time_start = time_start
        self.data_list = data_list
        self.write_dir = write_dir
        self.parallelio = parallelio
        self.name = name

        self.handle_init(kw)


class PICMI_LabFrameParticleDiagnostic(_ClassWithInit):
    """
    Defines the particle diagnostics in the lab frame

    Parameters
    ----------
    grid: grid instance
        Grid object for the diagnostic

    num_snapshots: integer
        Number of lab frame snapshots to make

    dt_snapshots: float
        Time between each snapshot in lab frame

    species: species instance or list of species instances, optional
        Species to write out. If not specified, all species are written.
        Note that the name attribute must be defined for the species.

    data_list: list of strings, optional
        The data to be written out. Possible values 'position', 'momentum', 'weighting'.
        Defaults to the output list of the implementing code.

    time_start: float, default=0
        Time for the first snapshot in lab frame

    write_dir: string, optional
        Directory where data is to be written

    parallelio: bool, optional
        If set to True, particle diagnostics are dumped in parallel

    name: string, optional
        Sets the base name for the diagnostic output files
    """
    def __init__(self, grid, num_snapshots, dt_snapshots, data_list=None,
                 time_start = 0.,
                 species = None,
                 write_dir = None,
                 parallelio = None,
                 name = None,
                 **kw):

        if data_list is not None:
            assert isinstance(data_list, list), 'LabFrameParticleDiagnostic: data_list must be a list'

        self.grid = grid
        self.num_snapshots = num_snapshots
        self.dt_snapshots = dt_snapshots
        self.time_start = time_start
        self.species = species
        self.data_list = data_list
        self.write_dir = write_dir
        self.parallelio = parallelio
        self.name = name

        self.handle_init(kw)


"""Classes following the PICMI standard
These should be the base classes for Python implementation of the PICMI standard
"""
import math
import sys
import re

from .base import _ClassWithInit, _get_constants

# ---------------
# Physics objects
# ---------------

class PICMI_GaussianLaser(_ClassWithInit):
    """
    Specifies a Gaussian laser distribution.

    More precisely, the electric field **near the focal plane** is given by:

    .. math::

        E(\\boldsymbol{x},t) = a_0\\times E_0\,
        \exp\left( -\\frac{r^2}{w_0^2} - \\frac{(z-z_0-ct)^2}{c^2\\tau^2} \\right)
        \cos[ k_0( z - z_0 - ct ) - \phi_{cep} ]

    where :math:`k_0 = 2\pi/\\lambda_0` is the wavevector and where
    :math:`E_0 = m_e c^2 k_0 / q_e` is the field amplitude for :math:`a_0=1`.

    .. note::

        The additional terms that arise **far from the focal plane**
        (Gouy phase, wavefront curvature, ...) are not included in the above
        formula for simplicity, but are of course taken into account by
        the code, when initializing the laser pulse away from the focal plane.

    Parameters
    ----------
    wavelength: float
        Laser wavelength [m], defined as :math:`\\lambda_0` in the above formula

    waist: float
        Waist of the Gaussian pulse at focus [m], defined as :math:`w_0` in the above formula

    duration: float
        Duration of the Gaussian pulse [s], defined as :math:`\\tau` in the above formula

    propagation_direction: unit vector of length 3 of floats
        Direction of propagation [1]

    polarization_direction: unit vector of length 3 of floats
        Direction of polarization [1]

    focal_position: vector of length 3 of floats
        Position of the laser focus [m]

    centroid_position: vector of length 3 of floats
        Position of the laser centroid at time 0 [m]

    a0: float
        Normalized vector potential at focus
        Specify either a0 or E0 (E0 takes precedence).

    E0: float
        Maximum amplitude of the laser field [V/m]
        Specify either a0 or E0 (E0 takes precedence).

    phi0: float
        Carrier envelope phase (CEP) [rad]

    zeta: float
        Spatial chirp at focus (in the lab frame) [m.s]

    beta: float
        Angular dispersion at focus (in the lab frame) [rad.s]

    phi2: float
        Temporal chirp at focus (in the lab frame) [s^2]

    fill_in: bool, default=True
        Flags whether to fill in the empty spaced opened up when the grid moves

    name: string, optional
        Optional name of the laser
    """
    def __init__(self, wavelength, waist, duration,
                 propagation_direction,
                 polarization_direction,
                 focal_position,
                 centroid_position,
                 a0 = None,
                 E0 = None,
                 phi0 = None,
                 zeta = None,
                 beta = None,
                 phi2 = None,
                 name = None,
                 fill_in = True,
                 **kw):

        assert E0 is not None or a0 is not None, 'One of E0 or a0 must be speficied'

        k0 = 2.*math.pi/wavelength
        if E0 is None:
            E0 = a0*_get_constants().m_e*_get_constants().c**2*k0/_get_constants().q_e
        if a0 is None:
            a0 = E0/(_get_constants().m_e*_get_constants().c**2*k0/_get_constants().q_e)

        self.wavelength = wavelength
        self.k0 = k0
        self.waist = waist
        self.duration = duration
        self.focal_position = focal_position
        self.centroid_position = centroid_position
        self.propagation_direction = propagation_direction
        self.polarization_direction = polarization_direction
        self.a0 = a0
        self.E0 = E0
        self.phi0 = phi0
        self.zeta = zeta
        self.beta = beta
        self.phi2 = phi2
        self.name = name
        self.fill_in = fill_in

        self.handle_init(kw)


class PICMI_AnalyticLaser(_ClassWithInit):
    """
    Specifies a laser with an analytically described distribution

    Parameters
    ----------
    name=None: string, optional
        Optional name of the laser

    field_expression: string
        Analytic expression describing the electric field of the laser [V/m]
        Expression should be in terms of the position, 'X', 'Y', in the plane orthogonal
        to the propagation direction, and 't' the time. The expression should describe
        the full field, including the oscillitory component.
        Parameters can be used in the expression with the values given as keyword arguments.

    wavelength: float
        Laser wavelength.
        This should be built into the expression, but some codes require a specified value for numerical purposes.

    propagation_direction: unit vector of length 3 of floats
        Direction of propagation [1]

    polarization_direction: unit vector of length 3 of floats
        Direction of polarization [1]

    amax: float, optional
        Maximum normalized vector potential.
        Specify either amax or Emax (Emax takes precedence).
        This should be built into the expression, but some codes require a specified value for numerical purposes.

    Emax: float, optional
        Maximum amplitude of the laser field [V/m].
        Specify either amax or Emax (Emax takes precedence).
        This should be built into the expression, but some codes require a specified value for numerical purposes.

    fill_in: bool, default=True
        Flags whether to fill in the empty spaced opened up when the grid moves
    """
    def __init__(self, field_expression,
                 wavelength,
                 propagation_direction,
                 polarization_direction,
                 amax = None,
                 Emax = None,
                 name = None,
                 fill_in = True,
                 **kw):

        assert Emax is not None or amax is not None, 'One of Emax or amax must be speficied'

        k0 = 2.*math.pi/wavelength
        if Emax is None:
            Emax = amax*_get_constants().m_e*_get_constants().c**2*k0/_get_constants().q_e
        if amax is None:
            amax = Emax/(_get_constants().m_e*_get_constants().c**2*k0/_get_constants().q_e)

        self.wavelength = wavelength
        self.field_expression = field_expression
        self.k0 = k0
        self.propagation_direction = propagation_direction
        self.polarization_direction = polarization_direction
        self.amax = amax
        self.Emax = Emax
        self.name = name
        self.fill_in = fill_in

        self.field_expression = '{}'.format(field_expression).replace('\n', '')

        # --- Find any user defined keywords in the kw dictionary.
        # --- Save them and delete them from kw.
        # --- It's up to the code to make sure that all parameters
        # --- used in the expression are defined.
        self.user_defined_kw = {}
        for k in list(kw.keys()):
            if re.search(r'\b%s\b'%k, self.field_expression):
                self.user_defined_kw[k] = kw[k]
                del kw[k]

        self.handle_init(kw)


# ------------------
# Numeric Objects
# ------------------


class PICMI_LaserAntenna(_ClassWithInit):
    """
    Specifies the laser antenna injection method

    Parameters
    ----------
    position: vector of strings
        Position of antenna launching the laser [m]

    normal_vector: vector of strings, optional
        Vector normal to antenna plane, defaults to the laser direction
        of propagation [1]
    """
    def __init__(self, position, normal_vector=None, **kw):

        self.position = position
        self.normal_vector = normal_vector

        self.handle_init(kw)



from .version import __version__

from .base import *
from .diagnostics import *
from .fields import *
from .applied_fields import *
from .interactions import *
from .lasers import *
from .particles import *
from .simulation import *



"""Classes following the PICMI standard
These should be the base classes for Python implementation of the PICMI standard
The classes in this file are related to interactions (e.g. field ionization, collisions, QED)
"""

from .base import _ClassWithInit

class PICMI_FieldIonization(_ClassWithInit):
    """
    Field ionization on an ion species

    Parameters
    ----------
    model: string
        Ionization model, e.g. "ADK"

    ionized_species: species instance
        Species that is ionized

    product_species: species instance
        Species in which ionized electrons are stored.
    """
    def __init__(self, model, ionized_species, product_species, **kw):
        self.model = model
        self.ionized_species = ionized_species
        self.product_species = product_species

        self.handle_init(kw)




"""Classes following the PICMI standard
These should be the base classes for Python implementation of the PICMI standard
The classes in the file are all particle related
"""
import math
import sys
import re
import numpy as np

from .base import _ClassWithInit

# ---------------
# Physics objects
# ---------------


class PICMI_Species(_ClassWithInit):
    """
    Sets up the species to be simulated.
    The species charge and mass can be specified by setting the particle type or by setting them directly.
    If the particle type is specified, the charge or mass can be set to override the value from the type.

    Parameters
    ----------
    particle_type: string, optional
        A string specifying an elementary particle, atom, or other, as defined in
        the openPMD 2 species type extension, openPMD-standard/EXT_SpeciesType.md

    name: string, optional
        Name of the species

    method: {'Boris', 'Vay', 'Higuera-Cary', 'Li' , 'free-streaming', 'LLRK4'}
        The particle advance method to use. Code-specific method can be specified using 'other:<method>'. The default is code
        dependent.

        - 'Boris': Standard "leap-frog" Boris advance
        - 'Vay':
        - 'Higuera-Cary':
        - 'Li' :
        - 'free-streaming': Advance with no fields
        - 'LLRK4': Landau-Lifschitz radiation reaction formula with RK-4)

    charge_state: float, optional
        Charge state of the species (applies only to atoms) [1]

    charge: float, optional
        Particle charge, required when type is not specified, otherwise determined from type [C]

    mass: float, optional
        Particle mass, required when type is not specified, otherwise determined from type [kg]

    initial_distribution: distribution instance
        The initial distribution loaded at t=0. Must be one of the standard distributions implemented.

    density_scale: float, optional
        A scale factor on the density given by the initial_distribution

    particle_shape: {'NGP', 'linear', 'quadratic', 'cubic'}
        Particle shape used for deposition and gather.
        If not specified, the value from the `Simulation` object will be used.
        Other values maybe specified that are code dependent.
    """

    methods_list = ['Boris' , 'Vay', 'Higuera-Cary', 'Li', 'free-streaming', 'LLRK4']

    def __init__(self, particle_type=None, name=None, charge_state=None, charge=None, mass=None,
                 initial_distribution=None, particle_shape=None, density_scale=None, method=None, **kw):


        assert method is None or method in PICMI_Species.methods_list or method.startswith('other:'), \
            Exception('method must starts with either "other:", or be one of the following '+', '.join(PICMI_Species.methods_list))

        self.method = method
        self.particle_type = particle_type
        self.name = name
        self.charge = charge
        self.charge_state = charge_state
        self.mass = mass
        self.initial_distribution = initial_distribution
        self.particle_shape = particle_shape
        self.density_scale = density_scale

        self.interactions = []

        self.handle_init(kw)


class PICMI_MultiSpecies(_ClassWithInit):
    """
    INCOMPLETE: proportions argument is not implemented
    Multiple species that are initialized with the same distribution.
    Each parameter can be list, giving a value for each species, or a single value which is given to all species.
    The species charge and mass can be specified by setting the particle type or by setting them directly.
    If the particle type is specified, the charge or mass can be set to override the value from the type.

    Parameters
    ----------
    particle_types: list of strings, optional
        A string specifying an elementary particle, atom, or other, as defined in
        the openPMD 2 species type extension, openPMD-standard/EXT_SpeciesType.md

    names: list of strings, optional
        Names of the species

    charge_states: list of floats, optional
        Charge states of the species (applies only to atoms)

    charges: list of floats, optional
        Particle charges, required when type is not specified, otherwise determined from type [C]

    masses: list of floats, optional
        Particle masses, required when type is not specified, otherwise determined from type [kg]

    proportions: list of floats, optional
        Proportions of the initial distribution made up by each species

    initial_distribution: distribution instance
        Initial particle distribution, applied to all species

    particle_shape: {'NGP', 'linear', 'quadratic', 'cubic'}
        Particle shape used for deposition and gather.
        If not specified, the value from the `Simulation` object will be used.
        Other values maybe specified that are code dependent.
    """
    # --- Note to developer: This class attribute needs to be set to the Species class
    # --- defined in the codes PICMI implementation.
    Species_class = None

    def __init__(self, particle_types=None, names=None, charge_states=None, charges=None, masses=None,
                 proportions=None, initial_distribution=None, particle_shape=None,
                 **kw):

        self.particle_types = particle_types
        self.names = names
        self.charges = charges
        self.charge_states = charge_states
        self.masses = masses
        self.proportions = proportions
        self.initial_distribution = initial_distribution
        self.particle_shape = particle_shape

        self.nspecies = None
        self.check_nspecies(particle_types)
        self.check_nspecies(names)
        self.check_nspecies(charges)
        self.check_nspecies(charge_states)
        self.check_nspecies(masses)
        self.check_nspecies(proportions)

        # --- Create the instances of each species
        self.species_instances_list = []
        self.species_instances_dict = {}
        for i in range(self.nspecies):
            particle_type = self.get_input_item(particle_types, i)
            name = self.get_input_item(names, i)
            charge = self.get_input_item(charges, i)
            charge_state = self.get_input_item(charge_states, i)
            mass = self.get_input_item(masses, i)
            proportion = self.get_input_item(proportions, i)
            specie = PICMI_MultiSpecies.Species_class(particle_type = particle_type,
                                                      name = name,
                                                      charge = charge,
                                                      charge_state = charge_state,
                                                      mass = mass,
                                                      initial_distribution = initial_distribution,
                                                      density_scale = proportion)
            self.species_instances_list.append(specie)
            if name is not None:
                self.species_instances_dict[name] = specie

        self.handle_init(kw)

    def check_nspecies(self, var):
        if var is not None:
            try:
                nvars = len(var)
            except TypeError:
                nvars = 1
            assert self.nspecies is None or self.nspecies == nvars, Exception('All inputs must have the same length')
            self.nspecies = nvars

    def get_input_item(self, var, i):
        if var is None:
            return None
        else:
            try:
                len(var)
            except TypeError:
                return var
            else:
                return var[i]

    def __len__(self):
        return self.nspecies

    def __getitem__(self, key):
        if isinstance(key, str):
            return self.species_instances_dict[key]
        else:
            return self.species_instances_list[key]


class PICMI_GaussianBunchDistribution(_ClassWithInit):
    """
    Describes a Gaussian distribution of particles

    Parameters
    ----------
    n_physical_particles: integer
        Number of physical particles in the bunch

    rms_bunch_size: vector of length 3 of floats
        RMS bunch size at t=0 [m]

    rms_velocity: vector of length 3 of floats, default=[0.,0.,0.]
        RMS velocity spread at t=0 [m/s]

    centroid_position: vector of length 3 of floats, default=[0.,0.,0.]
        Position of the bunch centroid at t=0 [m]

    centroid_velocity: vector of length 3 of floats, default=[0.,0.,0.]
        Velocity (gamma*V) of the bunch centroid at t=0 [m/s]

    velocity_divergence: vector of length 3 of floats, default=[0.,0.,0.]
        Expansion rate of the bunch at t=0 [m/s/m]
    """
    def __init__(self,n_physical_particles, rms_bunch_size,
                 rms_velocity = [0.,0.,0.],
                 centroid_position = [0.,0.,0.],
                 centroid_velocity = [0.,0.,0.],
                 velocity_divergence = [0.,0.,0.],
                 **kw):
        self.n_physical_particles = n_physical_particles
        self.rms_bunch_size = rms_bunch_size
        self.rms_velocity = rms_velocity
        self.centroid_position = centroid_position
        self.centroid_velocity = centroid_velocity
        self.velocity_divergence = velocity_divergence

        self.handle_init(kw)


class PICMI_UniformDistribution(_ClassWithInit):
    """
    Describes a uniform density distribution of particles

    Parameters
    ----------
    density: float
        Physical number density [m^-3]

    lower_bound: vector of length 3 of floats, optional
        Lower bound of the distribution [m]

    upper_bound: vector of length 3 of floats, optional
        Upper bound of the distribution [m]

    rms_velocity: vector of length 3 of floats, default=[0.,0.,0.]
        Thermal velocity spread [m/s]

    directed_velocity: vector of length 3 of floats, default=[0.,0.,0.]
        Directed, average, proper velocity [m/s]

    fill_in: bool, optional
        Flags whether to fill in the empty spaced opened up when the grid moves
    """

    def __init__(self, density,
                 lower_bound = [None,None,None],
                 upper_bound = [None,None,None],
                 rms_velocity = [0.,0.,0.],
                 directed_velocity = [0.,0.,0.],
                 fill_in = None,
                 **kw):
        self.density = density
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.rms_velocity = rms_velocity
        self.directed_velocity = directed_velocity
        self.fill_in = fill_in

        self.handle_init(kw)


class PICMI_FoilDistribution(_ClassWithInit):
    """
    Describes a foil with optional exponential pre- and post-plasma ramps along the propagation direction.

    Parameters
    ----------
    density: float
        Physical number density [m^-3]

    front : float
        postion of front surface of foil [m]

    thickness: float >= 0
        thickness of the foil [m]

    exponential_pre_plasma_length: float > 0, optional
        length scale of expoential decay of pre-foil plasma density [m]

    exponential_pre_plasma_cutoff : float >= 0, optional
        cutoff length for exponential decay of pre-foil density [m]

    exponential_post_plasma_length: float > 0, optional
        length scale of expoential decay of post-foil plasma density [m]

    exponential_post_plasma_cutoff : float >= 0, optional
        cutoff length for exponential decay of post-foil density [m]

    lower_bound: vector of length 3 of floats, optional
        Lower bound of the distribution [m]

    upper_bound: vector of length 3 of floats, optional
        Upper bound of the distribution [m]

    rms_velocity: vector of length 3 of floats, default=[0.,0.,0.]
        Thermal velocity spread [m/s]

    directed_velocity: vector of length 3 of floats, default=[0.,0.,0.]
        Directed, average, proper velocity [m/s]

    fill_in: bool, optional
        Flags whether to fill in the empty spaced opened up when the grid moves
    """

    def __init__(self,
                 density,
                 front,
                 thickness,
                 rms_velocity = [0., 0., 0.],
                 directed_velocity = [0., 0., 0.],
                 lower_bound = [None,None,None],
                 upper_bound = [None,None,None],
                 exponential_pre_plasma_length = None,
                 exponential_pre_plasma_cutoff = None,
                 exponential_post_plasma_length = None,
                 exponential_post_plasma_cutoff = None,
                 fill_in = None,
                 **kw) :
        self.density = density
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.rms_velocity = rms_velocity
        self.directed_velocity = directed_velocity
        self.fill_in = fill_in

        self.front = front
        self.thickness = thickness
        self.exponential_pre_plasma_length = exponential_pre_plasma_length
        self.exponential_pre_plasma_cutoff = exponential_pre_plasma_cutoff
        self.exponential_post_plasma_length = exponential_post_plasma_length
        self.exponential_post_plasma_cutoff = exponential_post_plasma_cutoff

        self.handle_init(kw)


class PICMI_UniformFluxDistribution(_ClassWithInit):
    """
    Describes a flux of particles following a provided analytic expression

    Parameters
    ----------
    flux: float
        Flux of particles [m^-2.s^-1]

    flux_normal_axis: string
        x, y, or z for 3D, x or z for 2D, or r, t, or z in RZ geometry

    surface_flux_position: double
        location of the injection plane [m] along the direction
        specified by `flux_normal_axis`

    flux_direction: int
        Direction of the flux relative to the plane: -1 or +1

    lower_bound: vector of floats, optional
        Lower bound of the distribution [m]

    upper_bound: vector of floats, optional
        Upper bound of the distribution [m]

    rms_velocity: vector of floats, default=[0.,0.,0.]
        Thermal velocity spread [m/s]

    directed_velocity: vector of floats, default=[0.,0.,0.]
        Directed, average, proper velocity [m/s]

    flux_tmin: float, optional
        Time at which the flux injection will be turned on.

    flux_tmax: float, optional
        Time at which the flux injection will be turned off.

    gaussian_flux_momentum_distribution: bool, optional
        If True, the momentum distribution is v*Gaussian,
        in the direction normal to the plane. Otherwise,
        the momentum distribution is simply Gaussian.
    """

    def __init__(self, flux, flux_normal_axis,
                 surface_flux_position, flux_direction,
                 lower_bound = [None,None,None],
                 upper_bound = [None,None,None],
                 rms_velocity = [0.,0.,0.],
                 directed_velocity = [0.,0.,0.],
                 flux_tmin = None,
                 flux_tmax = None,
                 gaussian_flux_momentum_distribution = None,
                 **kw):
        self.flux = flux
        self.flux_normal_axis = flux_normal_axis
        self.surface_flux_position = surface_flux_position
        self.flux_direction = flux_direction
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.rms_velocity = rms_velocity
        self.directed_velocity = directed_velocity
        self.flux_tmin = flux_tmin
        self.flux_tmax = flux_tmax
        self.gaussian_flux_momentum_distribution = gaussian_flux_momentum_distribution

        self.handle_init(kw)

class PICMI_AnalyticFluxDistribution(_ClassWithInit):
    """
    Describes a flux of particles emitted from a plane

    Parameters
    ----------
    flux: string
        Analytic expression describing flux of particles [m^-2.s^-1]
        Expression should be in terms of the position and time, written as 'x', 'y', 'z', and 't'.

    flux_normal_axis: string
        x, y, or z for 3D, x or z for 2D, or r, t, or z in RZ geometry

    surface_flux_position: double
        location of the injection plane [m] along the direction
        specified by `flux_normal_axis`

    flux_direction: int
        Direction of the flux relative to the plane: -1 or +1

    lower_bound: vector of floats, optional
        Lower bound of the distribution [m]

    upper_bound: vector of floats, optional
        Upper bound of the distribution [m]

    rms_velocity: vector of floats, default=[0.,0.,0.]
        Thermal velocity spread [m/s]

    directed_velocity: vector of floats, default=[0.,0.,0.]
        Directed, average, proper velocity [m/s]

    flux_tmin: float, optional
        Time at which the flux injection will be turned on.

    flux_tmax: float, optional
        Time at which the flux injection will be turned off.

    gaussian_flux_momentum_distribution: bool, optional
        If True, the momentum distribution is v*Gaussian,
        in the direction normal to the plane. Otherwise,
        the momentum distribution is simply Gaussian.
    """

    def __init__(self, flux_expression, flux_normal_axis,
                 surface_flux_position, flux_direction,
                 lower_bound = [None,None,None],
                 upper_bound = [None,None,None],
                 rms_velocity = [0.,0.,0.],
                 directed_velocity = [0.,0.,0.],
                 flux_tmin = None,
                 flux_tmax = None,
                 gaussian_flux_momentum_distribution = None,
                 **kw):
        self.flux_expression = f'{flux_expression}'.replace('\n', '')
        self.flux_normal_axis = flux_normal_axis
        self.surface_flux_position = surface_flux_position
        self.flux_direction = flux_direction
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.rms_velocity = rms_velocity
        self.directed_velocity = directed_velocity
        self.flux_tmin = flux_tmin
        self.flux_tmax = flux_tmax
        self.gaussian_flux_momentum_distribution = gaussian_flux_momentum_distribution

        self.user_defined_kw = {}
        for k in list(kw.keys()):
            if re.search(r'\b%s\b'%k, self.flux_expression):
                self.user_defined_kw[k] = kw[k]
                del kw[k]

        self.handle_init(kw)

class PICMI_AnalyticDistribution(_ClassWithInit):
    """
    Describes a plasma with density following a provided analytic expression

    Parameters
    ----------
    density_expression: string
        Analytic expression describing physical number density (string) [m^-3].
        Expression should be in terms of the position, written as 'x', 'y', and 'z'.
        Parameters can be used in the expression with the values given as keyword arguments.

    momentum_expressions: list of strings
        Analytic expressions describing the gamma*velocity for each axis [m/s].
        Expressions should be in terms of the position, written as 'x', 'y', and 'z'.
        Parameters can be used in the expression with the values given as keyword arguments.
        For any axis not supplied (set to None), directed_velocity will be used.

    momentum_spread_expressions: list of strings
        Analytic expressions describing the gamma*velocity Gaussian thermal spread sigma for each axis [m/s].
        Expressions should be in terms of the position, written as 'x', 'y', and 'z'.
        Parameters can be used in the expression with the values given as keyword arguments.
        For any axis not supplied (set to None), zero will be used.

    lower_bound: vector of length 3 of floats, optional
        Lower bound of the distribution [m]

    upper_bound: vector of length 3 of floats, optional
        Upper bound of the distribution [m]

    rms_velocity: vector of length 3 of floats, detault=[0.,0.,0.]
        Thermal velocity spread [m/s]

    directed_velocity: vector of length 3 of floats, detault=[0.,0.,0.]
        Directed, average, proper velocity [m/s]

    fill_in: bool, optional
        Flags whether to fill in the empty spaced opened up when the grid moves


    This example will create a distribution where the density is n0 below rmax and zero elsewhere.::

    .. code-block:: python

      dist = AnalyticDistribution(density_expression='((x**2+y**2)<rmax**2)*n0',
                                  rmax = 1.,
                                  n0 = 1.e20,
                                  ...)

    """

    def __init__(self, density_expression,
                 momentum_expressions = [None, None, None],
                 momentum_spread_expressions = [None, None, None],
                 lower_bound = [None,None,None],
                 upper_bound = [None,None,None],
                 rms_velocity = [0.,0.,0.],
                 directed_velocity = [0.,0.,0.],
                 fill_in = None,
                 **kw):
        self.density_expression = f'{density_expression}'.replace('\n', '')
        self.momentum_expressions = momentum_expressions
        self.momentum_spread_expressions = momentum_spread_expressions
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.rms_velocity = rms_velocity
        self.directed_velocity = directed_velocity
        self.fill_in = fill_in

        # --- Convert momentum expressions to string if needed.
        for idir in range(3):
            if self.momentum_expressions[idir] is not None:
                self.momentum_expressions[idir] = f'{self.momentum_expressions[idir]}'.replace('\n', '')
            if self.momentum_spread_expressions[idir] is not None:
                self.momentum_spread_expressions[idir] = f'{self.momentum_spread_expressions[idir]}'.replace('\n', '')

        # --- Find any user defined keywords in the kw dictionary.
        # --- Save them and delete them from kw.
        # --- It's up to the code to make sure that all parameters
        # --- used in the expression are defined.
        self.user_defined_kw = {}
        for k in list(kw.keys()):
            if re.search(r'\b%s\b'%k, self.density_expression):
                self.user_defined_kw[k] = kw[k]
                del kw[k]
            elif self.momentum_expressions[0] is not None and re.search(r'\b%s\b'%k, self.momentum_expressions[0]):
                self.user_defined_kw[k] = kw[k]
                del kw[k]
            elif self.momentum_expressions[1] is not None and re.search(r'\b%s\b'%k, self.momentum_expressions[1]):
                self.user_defined_kw[k] = kw[k]
                del kw[k]
            elif self.momentum_expressions[2] is not None and re.search(r'\b%s\b'%k, self.momentum_expressions[2]):
                self.user_defined_kw[k] = kw[k]
                del kw[k]

        self.handle_init(kw)


class PICMI_ParticleListDistribution(_ClassWithInit):
    """
    Load particles at the specified positions and velocities

    Parameters
    ----------
    x: float, default=0.
        List of x positions of the particles [m]

    y: float, default=0.
        List of y positions of the particles [m]

    z: float, default=0.
        List of z positions of the particles [m]

    ux: float, default=0.
        List of ux positions of the particles (ux = gamma*vx) [m/s]

    uy: float, default=0.
        List of uy positions of the particles (uy = gamma*vy) [m/s]

    uz: float, default=0.
        List of uz positions of the particles (uz = gamma*vz) [m/s]

    weight: float
        Particle weight or list of weights, number of real particles per simulation particle
    """
    def __init__(self, x=0., y=0., z=0., ux=0., uy=0., uz=0., weight=0.,
                 **kw):
        # --- Get length of arrays, set to one for scalars
        lenx = np.size(x)
        leny = np.size(y)
        lenz = np.size(z)
        lenux = np.size(ux)
        lenuy = np.size(uy)
        lenuz = np.size(uz)
        lenw = np.size(weight)

        maxlen = max(lenx, leny, lenz, lenux, lenuy, lenuz, lenw)
        assert lenx==maxlen or lenx==1, "Length of x doesn't match len of others"
        assert leny==maxlen or leny==1, "Length of y doesn't match len of others"
        assert lenz==maxlen or lenz==1, "Length of z doesn't match len of others"
        assert lenux==maxlen or lenux==1, "Length of ux doesn't match len of others"
        assert lenuy==maxlen or lenuy==1, "Length of uy doesn't match len of others"
        assert lenuz==maxlen or lenuz==1, "Length of uz doesn't match len of others"
        assert lenw==maxlen or lenw==1, "Length of weight doesn't match len of others"

        if lenx == 1:
            x = np.array(x)*np.ones(maxlen)
        if leny == 1:
            y = np.array(y)*np.ones(maxlen)
        if lenz == 1:
            z = np.array(z)*np.ones(maxlen)
        if lenux == 1:
            ux = np.array(ux)*np.ones(maxlen)
        if lenuy == 1:
            uy = np.array(uy)*np.ones(maxlen)
        if lenuz == 1:
            uz = np.array(uz)*np.ones(maxlen,'d')
        # --- Note that weight can be a scalar

        self.weight = weight
        self.x = x
        self.y = y
        self.z = z
        self.ux = ux
        self.uy = uy
        self.uz = uz

        self.handle_init(kw)


# ------------------
# Numeric Objects
# ------------------


class PICMI_ParticleDistributionPlanarInjector(_ClassWithInit):
    """
    Describes the injection of particles from a plane

    Parameters
    ----------
    position: vector of length 3 of floats
        Position of the particle centroid [m]

    plane_normal: vector of length 3 of floats
        Vector normal to the plane of injection [1]

    plane_velocity: vector of length 3 of floats
        Velocity of the plane of injection [m/s]

    method: {'InPlace', 'Plane'}
    """
    def __init__(self, position, plane_normal, plane_velocity=[0.,0.,0.], method='InPlace', **kw):
        self.position = position
        self.plane_normal = plane_normal
        self.plane_velocity = plane_velocity
        self.method = method

        self.handle_init(kw)


class PICMI_GriddedLayout(_ClassWithInit):
    """
    Specifies a gridded layout of particles

    Parameters
    ----------
    n_macroparticle_per_cell: vector of integers
        Number of particles per cell along each axis

    grid: grid instance, optional
        Grid object specifying the grid to follow.
        If not specified, the underlying grid of the code is used.
    """
    def __init__(self, n_macroparticle_per_cell, grid=None, **kw):
        self.n_macroparticle_per_cell = n_macroparticle_per_cell
        self.grid = grid

        self.handle_init(kw)


class PICMI_PseudoRandomLayout(_ClassWithInit):
    """
    Specifies a pseudo-random layout of the particles

    Parameters
    ----------
    n_macroparticles: integer
        Total number of macroparticles to load.
        Either this argument or n_macroparticles_per_cell should be supplied.

    n_macroparticles_per_cell: integer
        Number of macroparticles to load per cell.
        Either this argument or n_macroparticles should be supplied.

    seed: integer, optional
        Pseudo-random number generator seed

    grid: grid instance, optional
        Grid object specifying the grid to follow for n_macroparticles_per_cell.
        If not specified, the underlying grid of the code is used.
    """
    def __init__(self, n_macroparticles=None, n_macroparticles_per_cell=None, seed=None, grid=None, **kw):

        assert (n_macroparticles is not None)^(n_macroparticles_per_cell is not None), \
               Exception('Only one of n_macroparticles and n_macroparticles_per_cell must be specified')

        self.n_macroparticles = n_macroparticles
        self.n_macroparticles_per_cell = n_macroparticles_per_cell
        self.seed = seed
        self.grid = grid

        self.handle_init(kw)
